from queue import Queue, Empty
from System import Decimal as CSDecimal
from datetime import datetime

from IntelligencePy import Security_OrdType
from IntelligencePy import Security_Lot
from IntelligencePy import Security_Class
from IntelligencePy import Security_PriceFlag
from IntelligencePy import SIDE_FLAG
from IntelligencePy import TIME_IN_FORCE

from PackagePy import PT04010, PT06002

class Order:
    def __init__(self, RequestIds = dict(), callback = lambda dt, dic: print(dt, dic)):
        self.tradecom = None
        self.__RequestIdDict = RequestIds
        self.callback = callback

    def Order(self, ordtype: Security_OrdType, ordlot: Security_Lot, ordclass: Security_Class,
                BrokerId, Account, stockid, bs: SIDE_FLAG, qty, Price, 
                PF: Security_PriceFlag, subAccount, AgentID, OrderNo, TIF: TIME_IN_FORCE) -> PT04010:
        ordQty=int(qty)
        ordPrz=CSDecimal(float(Price))

        ##取得送單序號,以便回報時對應
        requestId=self.tradecom.GetRequestId()
        # requestId = Int64(rid)
        print("送單 RequestId=[{rid}]".format(rid=requestId))
        rtn = self.tradecom.SecurityOrder(
            requestId, 
            ordtype, 
            ordlot, 
            ordclass, 
            BrokerId, 
            Account, 
            stockid, 
            bs, 
            ordQty, 
            ordPrz, 
            PF, 
            subAccount, 
            AgentID, 
            OrderNo, 
            TIF
        )
        if (rtn != 0):
            print("SecurityOrder Error :"+str(rtn))
            print(self.tradecom.GetOrderErrMsg(rtn))
            raise SyntaxError(f"SecurityOrder Error : {str(rtn)} , {self.tradecom.GetOrderErrMsg(rtn)}")

        self.__RequestIdDict[requestId] = Queue()
        SECU_ORDER_RPT = self.getPkg(self.__RequestIdDict[requestId])
        if isinstance(SECU_ORDER_RPT, PT04010):
            return PT04010(SECU_ORDER_RPT)
        elif isinstance(SECU_ORDER_RPT, PT06002):
            SECU_ORDER_RPT.OrderNo = ""
            SECU_ORDER_RPT.CNTN = SECU_ORDER_RPT.CNT
            SECU_ORDER_RPT.ErrCode = SECU_ORDER_RPT.ErrorCode
            SECU_ORDER_RPT.ClientOrderTimeN = datetime.now().strftime("%Y%m%d%H%M%S%f")
            SECU_ORDER_RPT.Qty = 0
            return SECU_ORDER_RPT
        else:
            return None



from IntelligencePy import DT
from PackagePy import *

class Report:
    def __init__(self,
                 RequestIds = dict(),
                 CNTs = dict(),
                 ORDER_RPTs = dict(),
                 DEAL_RPTs = dict(),
                 callback = lambda dt, dic: print(dt, dic)):
        
        self.callback = callback        
        self.__RequestIdDict = RequestIds
        self.__CNTDict = CNTs
        self.__ORDER_RPT_List = ORDER_RPTs
        self.__DEAL_RPT_List = DEAL_RPTs

    def Future_RcvMessage(self, pkgDT:DT, pkg):
        if pkgDT == DT.FUT_ORDER_ACK:
            FUT_ORDER_ACK = PT02002(pkg)
            self.callback(pkgDT, FUT_ORDER_ACK)
            if FUT_ORDER_ACK.ErrorCode != 0:
                FUT_ORDER_ACK.ErrMsg = self.tradecom.GetMessageMap(FUT_ORDER_ACK.ErrorCode)
                self.__RequestIdDict[pkg.RequestId].put(FUT_ORDER_ACK)
            else:
                self.__CNTDict[pkg.CNT] = pkg.RequestId

        elif pkgDT == DT.FUT_ORDER_RPT:
            FUT_ORDER_RPT = PT02010(pkg)
            self.callback(pkgDT, FUT_ORDER_RPT)
            if self.__ORDER_RPT_List.get(FUT_ORDER_RPT.OrderNo) is None:
                self.__ORDER_RPT_List[FUT_ORDER_RPT.OrderNo] = [FUT_ORDER_RPT]
            else:
                self.__ORDER_RPT_List[FUT_ORDER_RPT.OrderNo].append(FUT_ORDER_RPT)
            
            RequestId = self.__CNTDict.pop(pkg.CNT, None)    #檢查是否來自ACK
            if RequestId : self.__RequestIdDict[RequestId].put(FUT_ORDER_RPT)

        elif pkgDT == DT.FUT_DEAL_RPT:
            FUT_DEAL_RPT = PT02011(pkg)
            self.callback(pkgDT, FUT_DEAL_RPT)
            if self.__DEAL_RPT_List.get(FUT_DEAL_RPT.OrderNo) is None:
                self.__DEAL_RPT_List[FUT_DEAL_RPT.OrderNo] = [FUT_DEAL_RPT]
            else:
                self.__DEAL_RPT_List[FUT_DEAL_RPT.OrderNo].append(FUT_DEAL_RPT)
        
        else:
            return False
        return True


from IntelligencePy import DT
from PackagePy import *

class Report:
    def __init__(self,
                 RequestIds = dict(),
                 CNTs = dict(),
                 ORDER_RPTs = dict(),
                 DEAL_RPTs = dict(),
                 callback = lambda dt, dic: print(dt, dic)):
        
        self.callback = callback        
        self.__RequestIdDict = RequestIds
        self.__CNTDict = CNTs
        self.__ORDER_RPT_List = ORDER_RPTs
        self.__DEAL_RPT_List = DEAL_RPTs
    
    def Stock_RcvMessage(self, pkgDT:DT, pkg):
        if pkgDT == DT.SECU_ORDER_ACK_N:
            SECU_ORDER_ACK_N = PT06002(pkg)
            self.callback(pkgDT, SECU_ORDER_ACK_N)
            if SECU_ORDER_ACK_N.ErrorCode != 0:
                SECU_ORDER_ACK_N.ErrMsg=self.tradecom.GetMessageMap(SECU_ORDER_ACK_N.ErrorCode)
                self.__RequestIdDict[pkg.RequestId].put(SECU_ORDER_ACK_N)
            else:
                self.__CNTDict[pkg.CNT] = pkg.RequestId

        elif pkgDT == DT.SECU_ORDER_RPT:
            SECU_ORDER_RPT = PT04010(pkg)
            self.callback(pkgDT, SECU_ORDER_RPT)
            if self.__ORDER_RPT_List.get(SECU_ORDER_RPT.OrderNo) is None:
                self.__ORDER_RPT_List[SECU_ORDER_RPT.OrderNo] = [SECU_ORDER_RPT]
            else:
                self.__ORDER_RPT_List[SECU_ORDER_RPT.OrderNo].append(SECU_ORDER_RPT)
            
            RequestId = self.__CNTDict.pop(pkg.CNTN, None)    #檢查是否來自ACK
            if RequestId : self.__RequestIdDict[RequestId].put(SECU_ORDER_RPT)

        elif pkgDT == DT.SECU_DEAL_RPT:
            SECU_DEAL_RPT = PT04011(pkg)
            self.callback(pkgDT, SECU_DEAL_RPT)
            if self.__DEAL_RPT_List.get(SECU_DEAL_RPT.OrderNo) is None:
                self.__DEAL_RPT_List[SECU_DEAL_RPT.OrderNo] = [SECU_DEAL_RPT]
            else:
                self.__DEAL_RPT_List[SECU_DEAL_RPT.OrderNo].append(SECU_DEAL_RPT)
        else:
            return False
        return True


import sys, os
import clr
from queue import Queue, Empty
import datetime, time
import json
from System import Decimal as CSDecimal

assembly_path = os.path.dirname(__file__)
sys.path.append(assembly_path)
clr.AddReference("TradeCom")     #必要引用dll

from Smart import TaiFexCom
from IntelligencePy import DT
from IntelligencePy import COM_STATUS
from IntelligencePy import RECOVER_STATUS

from PackagePy import *
from Login import Login
from Stock.Order import Order
from Stock.Report import Report
from Future.Order import Order as FutOrder
from Future.Report import Report as FutReport

class TradeCom(Login, Order, Report, FutOrder, FutReport):
    def __init__(self, host, port, sid, timeout=5000, callback = lambda dt, dic: print(dt, dic)) -> None:
        """程式初始化

        Args:
            host (str): 主機連線的host
            port (num): 主機連線的port
            sid (str):  主機連線的sid
        """
        self.__RequestIdDict = dict()
        self.__CNTDict = dict()
        self.__ORDER_RPT_List = dict()
        self.__DEAL_RPT_List = dict()

        Login.__init__(self, host, port)
        Order.__init__(self, self.__RequestIdDict)
        Report.__init__(self, self.__RequestIdDict, self.__CNTDict, self.__ORDER_RPT_List, self.__DEAL_RPT_List)
        FutOrder.__init__(self, self.__RequestIdDict)
        FutReport.__init__(self, self.__RequestIdDict, self.__CNTDict, self.__ORDER_RPT_List, self.__DEAL_RPT_List)

        self.sid = sid
        self.callback = callback
        self.tradecom = TaiFexCom("", port, sid)
        self.tradecom.ConnectTimeout = timeout
        print("TradeCom API 初始化 Version (%s) ........" % (self.tradecom.version))
        # register event handler
        #狀態通知事件KGI Tradecom API message event
        self.tradecom.OnRcvMessage += self.onTradeRcvMessage
        #資料接收事件KGI Tradecom API status event
        self.tradecom.OnGetStatus += self.onTradeGetStatus
        #資料回補事件KGI Tradecom API Recover event
        self.tradecom.OnRecoverStatus += self.onTradeRecoverStatus
        #資料回補事件KGI Tradecom API Server Time event
        self.tradecom.OnRcvServerTime += self.onTradeRcvServerTime

        #是否回下載商品檔
        # self.tradecom.AutoRetriveProductInfo=True

        self.debug = True   #正式/模擬環境

        self.com_status = None
        self.Com_StatusDict = dict()
        self.GetStatusQueue = Queue()
        self.RcvMessageQueue = Queue()
        # self.RecoverTopic = list()
    
    #region 回報
    def onTradeRcvMessage(self, sender, pkg): 
        try:
            pkgDT = DT(pkg.DT)
            if self.Login_RcvMessage(pkgDT, pkg):                
                pass
            elif self.Stock_RcvMessage(pkgDT, pkg):
                pass
            elif self.Future_RcvMessage(pkgDT, pkg):
                pass
            else:
                self.callback(pkgDT, PackageBase(pkg))
                pass
        except Exception as e:
            print('Exception.RcvMessage', e)  

    def onTradeGetStatus(self, sender, status, msg):
        try:            
            self.com_status = COM_STATUS(status)
            smsg = bytes(msg).decode('UTF-8','strict')
            self.Com_StatusDict[self.com_status] = { 'Time' : datetime.datetime.now() , 'Msg' : smsg }
            # print('onTradeGetStatus:[' + str(status) + ']:' + smsg)
            self.Login_GetStatus(self.com_status, smsg)
        except Exception as e:
            print('Exception.GetStatus'+e)  

    def onTradeRecoverStatus(self, sender, topic, status, count):
        try:
            status = RECOVER_STATUS(status)
            if (status==RECOVER_STATUS.RS_DONE):
                #回補資料結束
                if (count==0):
                    print("結束回補 Topic:["+topic+"]")
                else: 
                    print("結束回補 Topic=[{tp}], 筆數=[{count}]".format(tp=topic, count=count))
                # self.RecoverTopic.append(topic)
            elif (status==RECOVER_STATUS.RS_BEGIN):
                #開始回補資料
                print("開始回補 Topic:["+topic+"]")
            else:
                print("OnRecoverStatus Topic=[{tp}], Status=[{status}], 筆數=[{count}]".format(tp=topic, status=status, count=count))
        except Exception as e:
            print(e)  

    def onTradeRcvServerTime(self, sender, time, quality):
        try:
            # print('onTradeRcvServerTime', time, quality)
            self.callback(DT.NOTICE, f'onTradeRcvServerTime, {time}, {quality}')
        except Exception as e:
            print(e)  
    #endregion

    def getPkg(self, queue, timeout: float = 3):
        try:            
            pkg = queue.get(timeout=timeout)            
        except Empty:
            print('Timeout')
            return None
        
        return pkg
    
    @property
    def ORDER_RPT_List(self):
        return self.__ORDER_RPT_List
    
    @property
    def DEAL_RPT_List(self):
        return self.__DEAL_RPT_List


from enum import Enum
from IntelligencePy import *

ACTION_BUY = "Buy"
ACTION_SELL = "Sell"

STOCK_PRICE_TYPE_LIMITPRICE = "LMT"
STOCK_PRICE_TYPE_MKT = "MKT"
STOCK_PRICE_TYPE_CLOSE = "Close"

STOCK_ORDER_LOT_COMMON = "Common"  # 整股
STOCK_ORDER_LOT_BLOCKTRADE = "BlockTrade"  # 鉅額
STOCK_ORDER_LOT_FIXING = "Fixing"  # 定盤
STOCK_ORDER_LOT_ODD = "Odd"  # 零股
STOCK_ORDER_LOT_INTRADAY_ODD = "IntradayOdd"  # 零股

STOCK_ORDER_TYPE_COMMON = "Common"  # 整股
STOCK_ORDER_TYPE_BLOCKTRADE = "BlockTrade"  # 鉅額
STOCK_ORDER_TYPE_FIXING = "Fixing"  # 定盤
STOCK_ORDER_TYPE_ODD = "Odd"  # 零股

STOCK_ORDER_COND_CASH = "Cash"  # 現股
STOCK_ORDER_COND_NETTING = "Netting"  # 餘額交割
STOCK_ORDER_COND_MARGINTRADING = "MarginTrading"  # 融資
STOCK_ORDER_COND_SHORTSELLING = "ShortSelling"  # 融券
STOCK_ORDER_COND_EMERGING = "Emerging"  # 興櫃

STOCK_FIRST_SELL_YES = "true"
STOCK_FIRST_SELL_NO = "false"

FUTURES_PRICE_TYPE_LMT = "LMT"
FUTURES_PRICE_TYPE_MKT = "MKT"
FUTURES_PRICE_TYPE_MKP = "MKP"

ORDER_TYPE_ROD = "ROD"
ORDER_TYPE_IOC = "IOC"
ORDER_TYPE_FOK = "FOK"

FUTURES_OCTYPE_AUTO = "Auto"
FUTURES_OCTYPE_NEWPOSITION = "New"
FUTURES_OCTYPE_COVER = "Cover"
FUTURES_OCTYPE_DAYTRADE = "DayTrade"

FUTURES_CALLPUT_FUT = "F"
FUTURES_CALLPUT_CALL = "C"
FUTURES_CALLPUT_PUT = "P"

QUOTE_TYPE_TICK = "tick"
QUOTE_TYPE_BIDASK = "bidask"
QUOTE_TYPE_QUOTE = "quote"


class Action(str, Enum):
    Buy = ACTION_BUY
    Sell = ACTION_SELL

    def MapTo(self):
        return {
            Action.Buy : SIDE_FLAG.SF_BUY,
            Action.Sell : SIDE_FLAG.SF_SELL,
        }.get(self, None)

class OrderType(str, Enum):
    ROD = ORDER_TYPE_ROD
    IOC = ORDER_TYPE_IOC
    FOK = ORDER_TYPE_FOK

    def MapTo(self):
        return {
            OrderType.ROD : TIME_IN_FORCE.TIF_ROD,
            OrderType.IOC : TIME_IN_FORCE.TIF_IOC,
            OrderType.FOK : TIME_IN_FORCE.TIF_FOK
        }.get(self, None)

class StockPriceType(str, Enum):
    LMT = STOCK_PRICE_TYPE_LIMITPRICE
    MKT = STOCK_PRICE_TYPE_MKT
    FallStopPrice = Security_PriceFlag.SP_FallStopPrice
    UnchangePrice = Security_PriceFlag.SP_UnchangePrice
    RiseStopPrice = Security_PriceFlag.SP_RiseStopPrice

    def MapTo(self):
        return {
            StockPriceType.LMT : Security_PriceFlag.SP_FixedPrice,  #限價
            StockPriceType.MKT : Security_PriceFlag.SP_MarketPrice, #市價
            StockPriceType.FallStopPrice : Security_PriceFlag.SP_FallStopPrice, #跌停
            StockPriceType.UnchangePrice : Security_PriceFlag.SP_UnchangePrice,  #平盤
            StockPriceType.RiseStopPrice : Security_PriceFlag.SP_RiseStopPrice   #漲停
        }.get(self, None)

class StockOrderLot(str, Enum):
    Common = STOCK_ORDER_LOT_COMMON  # 整股
    BlockTrade = STOCK_ORDER_LOT_BLOCKTRADE  # 鉅額
    Fixing = STOCK_ORDER_LOT_FIXING  # 定盤
    Odd = STOCK_ORDER_LOT_ODD  # 零股
    IntradayOdd = STOCK_ORDER_LOT_INTRADAY_ODD  # 盤中零股

    def MapTo(self):
        return {
            StockOrderLot.Common : Security_Lot.Even_Lot,
            StockOrderLot.Odd : Security_Lot.Odd_Lot,
            StockOrderLot.Fixing : Security_Lot.Fixed_Price,
            StockOrderLot.BlockTrade : Security_Lot.Block_Trade,
            StockOrderLot.IntradayOdd : Security_Lot.Odd_InTraday,
        }.get(self, None)
       

class StockOrderCond(str, Enum):
    Cash = STOCK_ORDER_COND_CASH  # 現股
    # Netting = STOCK_ORDER_COND_NETTING  # 餘額交割
    MarginTrading = STOCK_ORDER_COND_MARGINTRADING  # 融資
    ShortSelling = STOCK_ORDER_COND_SHORTSELLING  # 融券
    # Emerging = STOCK_ORDER_COND_EMERGING  # 興櫃
    ShortUnLimit = Security_Class.SC_ShortUnLimit   #借券賣出
    EDN = Security_Class.SC_EDN #賣出擔保品

    def MapTo(self):
        return {
            StockOrderCond.Cash : Security_Class.SC_Ordinary,
            StockOrderCond.MarginTrading : Security_Class.SC_SelfMargin,
            StockOrderCond.ShortSelling : Security_Class.SC_SelfShort,
            StockOrderCond.ShortUnLimit : Security_Class.SC_ShortUnLimit,
            StockOrderCond.EDN : Security_Class.SC_EDN
        }.get(self, None)

class FuturesPriceType(str, Enum):
    LMT = FUTURES_PRICE_TYPE_LMT
    MKT = FUTURES_PRICE_TYPE_MKT
    MKP = FUTURES_PRICE_TYPE_MKP

    def MapTo(self):
        return {
            FuturesPriceType.LMT : PRICE_FLAG.PF_SPECIFIED,
            FuturesPriceType.MKT : PRICE_FLAG.PF_MARKET,
            FuturesPriceType.MKP : PRICE_FLAG.PF_MARKET_RANGE,
        }.get(self, None)


class FuturesOCType(str, Enum):
    Auto = FUTURES_OCTYPE_AUTO
    New = FUTURES_OCTYPE_NEWPOSITION
    Cover = FUTURES_OCTYPE_COVER
    DayTrade = FUTURES_OCTYPE_DAYTRADE

    def MapTo(self):
        return {
            FuturesOCType.Auto : POSITION_EFFECT.PE_AUTO,
            FuturesOCType.New : POSITION_EFFECT.PE_OPEN,
            FuturesOCType.Cover : POSITION_EFFECT.PE_CLOSE,
            FuturesOCType.DayTrade : POSITION_EFFECT.PE_DAY_TRADE
        }.get(self, None)

class SecurityType(str, Enum):
    Index = "IND"
    Stock = "STK"
    Future = "FUT"
    Option = "OPT"


class Exchange(str, Enum):
    TSE = "TSE"
    OTC = "OTC"
    OES = "OES"
    TAIFEX = "TAIFEX"


class Currency(str, Enum):
    TWD = "TWD"
    USD = "USD"
    HKD = "HKD"
    GBP = "GBP"
    AUD = "AUD"
    CAD = "CAD"
    SGD = "SGD"
    CHF = "CHF"
    JPY = "JPY"
    ZAR = "ZAR"
    SEK = "SEK"
    NZD = "NZD"
    THB = "THB"
    PHP = "PHP"
    IDR = "IDR"
    EUR = "EUR"
    KRW = "KRW"
    VND = "VND"
    MYR = "MYR"
    CNY = "CNY"


class OptionRight(str, Enum):
    No = ""
    Call = "C"
    Put = "P"


class Status(str, Enum):
    Cancelled = "Cancelled"
    Filled = "Filled"
    # Filling = 'Filling'
    PartFilled = "PartFilled"
    Inactive = "Inactive"
    Failed = "Failed"
    PendingSubmit = "PendingSubmit"
    PreSubmitted = "PreSubmitted"
    Submitted = "Submitted"
    # _DoneStates = {'Cancelled', 'Filled', 'Inactive', 'Failed'}
    # _ActiveStates = {'PendingSubmit', 'PreSubmitted', 'Submitted', 'PartFilled'}


class OrderState(str, Enum):
    StockDeal = "SDEAL"
    StockOrder = "SORDER"
    FuturesOrder = "FORDER"
    FuturesDeal = "FDEAL"


class QuoteType(str, Enum):
    Tick = QUOTE_TYPE_TICK
    BidAsk = QUOTE_TYPE_BIDASK
    Quote = QUOTE_TYPE_QUOTE


class QuoteVersion(str, Enum):
    v1 = "v1"


class DayTrade(str, Enum):
    Yes = "Yes"
    OnlyBuy = "OnlyBuy"
    No = "No"


class TickType(str, Enum):
    No = "None"  # 無法判斷0
    Buy = "Buy"  # 外盤1
    Sell = "Sell"  # 內盤2


class ChangeType(str, Enum):
    LimitUp = "LimitUp"  # 1漲停
    Up = "Up"  # 2漲
    Unchanged = "Unchanged"  # 3平盤
    Down = "Down"  # 4跌
    LimitDown = "LimitDown"  # 5跌停


class Unit(str, Enum):
    Common = "Common"
    Share = "Share"


class TradeType(str, Enum):
    Common = "Common"
    DayTrade = "DayTrade"


class ScannerType(str, Enum):
    ChangePercentRank = "ChangePercentRank"
    ChangePriceRank = "ChangePriceRank"
    DayRangeRank = "DayRangeRank"
    VolumeRank = "VolumeRank"
    AmountRank = "AmountRank"
    TickCountRank = "TickCountRank"


class TicksQueryType(str, Enum):
    AllDay = "AllDay"
    RangeTime = "RangeTime"
    LastCount = "LastCount"



import time
import typing
# from shioaji.base import BaseModel, StrictInt
from constant import (
    Action,
    Currency,
    Exchange,
    OptionRight,
    SecurityType,
    DayTrade,
)
from enum import Enum
from QuoteCom import Quote as quoteCom

__all__ = (
    "BaseContract",
    "Contract",
    "Index",
    "Stock",
    "Future",
    "Option",
    "Contracts",
    "ComboBase",
    "ComboContract",
)


class FetchStatus(str, Enum):
    Fetched = "Fetched"
    Fetching = "Fetching"
    Unfetch = "Unfetch"

#region Contracts

class BaseContract:
    security_type: SecurityType
    exchange: Exchange
    code: str

    def __str__(self):
        return str(self.__dict__)

    def astype(self):
        return _CONTRACTTYPE.get(self.security_type, self.__class__)(**self.dict())


class Contract(BaseContract):
    symbol: str = ""
    name: str = ""
    category: str = ""
    currency: Currency = Currency.TWD
    delivery_month: str = ""
    delivery_date: str = ""
    strike_price: typing.Union[float] = 0
    option_right: OptionRight = OptionRight.No
    underlying_kind: str = ""
    underlying_code: str = ""
    unit: typing.Union[float] = 0
    multiplier: int = 0
    limit_up: float = 0.0
    limit_down: float = 0.0
    reference: float = 0.0
    update_date: str = ""
    margin_trading_balance: int = 0
    short_selling_balance: int = 0
    day_trade: DayTrade = DayTrade.No
    target_code: str = ""


# class ComboBase(Contract):
#     action: Action


# class ComboContract(BaseModel):
#     legs: typing.List[ComboBase]


class Index(Contract):
    security_type: SecurityType = SecurityType.Index

    def __init__(self, **kwargs):
        self.exchange = kwargs.get("exchange", None)
        IDX = kwargs.get("IDX", None)
        if IDX:
            self.code = IDX.code
            self.symbol = f'{self.exchange}{self.code}'
            self.name = IDX.name

class Stock(Contract):
    security_type: SecurityType = SecurityType.Stock
    limit_up: float = 0.0
    limit_down: float = 0.0
    reference: float = 0.0
    margin_trading_balance: int = 0
    short_selling_balance: int = 0
    update_date: str = ""


class Future(Contract):
    security_type: SecurityType = SecurityType.Future
    exchange: Exchange = Exchange.TAIFEX
    limit_up: float = 0.0
    limit_down: float = 0.0
    reference: float = 0.0
    update_date: str = ""
    target_code: str = ""


class Option(Contract):
    security_type: SecurityType = SecurityType.Option
    exchange: Exchange = Exchange.TAIFEX
    limit_up: float = 0.0
    limit_down: float = 0.0
    reference: float = 0.0
    update_date: str = ""


_CONTRACTTYPE = {
    SecurityType.Index: Index,
    SecurityType.Stock: Stock,
    SecurityType.Future: Future,
    SecurityType.Option: Option,
}

#endregion

#region ProductContracts

# ProductTypeDict = dict(
#     IndexContracts="exchange",
#     StockContracts="exchange",
#     FutureContracts="category",
#     OptionContracts="category",
# )

# StreamProductTypeDict = dict(
#     StreamIndexContracts="exchange",
#     StreamStockContracts="exchange",
#     StreamFutureContracts="category",
#     StreamOptionContracts="category",
# )

# class BaseIterContracts:
#     def __iter__(self):
#         for key in self.__slots__:
#             if not key.startswith("_"):
#                 yield getattr(self, key)

#     def __bool__(self):
#         return True if list(self.keys()) else False

#     def keys(self):
#         return (key for key in self.__slots__ if not key.startswith("_"))

#     def __getitem__(self, key):
#         return getattr(self, key)

#     def get(self, key, default=None):
#         return getattr(self, key, default)

#     def __eq__(self, other: Contract) -> bool:
#         if isinstance(other, BaseModel):
#             return self.__dict__ == other.__dict__
#         else:
#             return self.__dict__ == other

# class ProductContracts(BaseIterContracts):
#     def __init__(self, contracts):
#         type_key = ProductTypeDict[self.__class__.__name__]
#         key_list = list(sorted(set([con[type_key] for con in contracts])))
#         self._code2contract = {}
#         self._fetched = False
#         self.__slots__ = key_list + ["_code2contract", "_fetched"]
#         if contracts:
#             for key in key_list:
#                 contract = [con for con in contracts if con[type_key] == key]
#                 mulcontract = MultiContract(key, contract)
#                 setattr(self, key, mulcontract)
#                 self._code2contract.update(getattr(self, key)._code2contract)
#         else:
#             self.__slots__ = ["_code2contract", "_fetched"]

#     def post_init(self, contracts: BaseIterContracts):
#         self.__slots__ = contracts.__slots__
#         for s in contracts.__slots__:
#             setattr(self, s, getattr(contracts, s))

#     def __repr__(self):
#         self._block()
#         return "({})".format(", ".join(self.__slots__[:-2]))

#     def __getitem__(self, key):
#         if not key.startswith("_"):
#             self._block()
#         return getattr(self, key, self._code2contract.get(key, None))

#     def get(self, key, default=None):
#         return getattr(self, key, self._code2contract.get(key, default))

#     def _block(self):
#         if not self._fetched:
#             for _ in range(100):
#                 time.sleep(0.3)
#                 if self._fetched:
#                     break

#     def __getattr__(self, attr):
#         if not attr.startswith("_"):
#             self._block()
#         if object.__getattribute__(self, attr):
#             return object.__getattribute__(self, attr)

#     def __getattribute__(self, item):
#         return super(ProductContracts, self).__getattribute__(item)


# class StreamProductContracts(ProductContracts):
#     def __init__(self, contracts):
#         type_key = StreamProductTypeDict[self.__class__.__name__]
#         key_list = list(sorted(set([con[type_key] for con in contracts])))
#         self._code2contract = {}
#         self._fetched = False
#         self._page_list = []
#         self._max_page = 0
#         self.__slots__ = key_list + [
#             "_code2contract",
#             "_fetched",
#             "_page_list",
#             "_max_page",
#         ]
#         if contracts:
#             for key in key_list:
#                 contract = [con for con in contracts if con[type_key] == key]
#                 mulcontract = StreamMultiContract(key, contract)
#                 setattr(self, key, mulcontract)
#                 self._code2contract.update(getattr(self, key)._code2contract)
#         else:
#             self.__slots__ = [
#                 "_code2contract",
#                 "_fetched",
#                 "_page_list",
#                 "_max_page",
#             ]

#     def __repr__(self):
#         self._block()
#         return "({})".format(", ".join(self.__slots__[:-4]))

#     def append(self, value):
#         slots = list(self.__slots__[:-4])
#         key_list = value.__slots__[:-4]
#         for key in key_list:
#             if key in slots:
#                 getattr(self, key).append(getattr(value, key))
#             else:
#                 mulcontract = getattr(value, key)
#                 setattr(self, key, mulcontract)
#                 self._code2contract.update(getattr(self, key)._code2contract)

#         slots.extend([key for key in key_list if key not in slots])
#         self.__slots__ = sorted(slots) + [
#             "_code2contract",
#             "_fetched",
#             "_page_list",
#             "_max_page",
#         ]
#         self._code2contract.update(value._code2contract)

#     def set_status_fetched(self):
#         self._fetched = True


# class IndexContracts(ProductContracts):
#     pass


# class StockContracts(ProductContracts):
#     pass


# class FutureContracts(ProductContracts):
#     pass


# class OptionContracts(ProductContracts):
#     pass


# class StreamIndexContracts(StreamProductContracts):
#     pass


# class StreamStockContracts(StreamProductContracts):
#     pass


# class StreamFutureContracts(StreamProductContracts):
#     pass


# class StreamOptionContracts(StreamProductContracts):
#     pass


# class MultiContract(BaseIterContracts):
#     def __init__(self, name, contracts):
#         self._name = name
#         self._code2contract = {}
#         self.__slots__ = []
#         for cont in contracts:
#             self.__slots__.append(cont["symbol"])
#             setattr(self, cont["symbol"], Contract(**cont).astype())
#             self._code2contract.update({cont["code"]: getattr(self, cont["symbol"])})
#         self.__slots__ += ["_name", "_code2contract"]

#     def __getitem__(self, key):
#         return getattr(self, key, self._code2contract.get(key, None))

#     def get(self, key, default=None):
#         return getattr(self, key, self._code2contract.get(key, default))

#     def __repr__(self):
#         return "{}({})".format(self._name, (", ").join(self.__slots__[:-2]))


# class StreamMultiContract(MultiContract):
#     def append(self, value):
#         slots = self.__slots__[:-2]
#         key_list = value.__slots__[:-2]
#         for key in key_list:
#             mulcontract = getattr(value, key)
#             setattr(self, key, mulcontract)

#         slots.extend(key_list)
#         self.__slots__ = slots + ["_name", "_code2contract"]
#         self._code2contract.update(value._code2contract)


# SecurityType2ProductContratcs = {
#     SecurityType.Index: StreamIndexContracts,
#     SecurityType.Stock: StreamStockContracts,
#     SecurityType.Future: StreamFutureContracts,
#     SecurityType.Option: StreamOptionContracts,
# }


# def get_product_contracts(security_type: SecurityType) -> ProductContracts:
#     return SecurityType2ProductContratcs.get(security_type, ProductContracts)


# class Contracts(BaseModel, BaseIterContracts):
#     Indexs: StreamIndexContracts
#     Stocks: StreamStockContracts
#     Futures: StreamFutureContracts
#     Options: StreamOptionContracts
#     status: FetchStatus = FetchStatus.Unfetch

#     class Config:
#         arbitrary_types_allowed = True

#     def __init__(self, **kwargs):
#         super().__init__(
#             **dict(
#                 Indexs=StreamIndexContracts(kwargs.get("Indexs", {})),
#                 Stocks=StreamStockContracts(kwargs.get("Stocks", {})),
#                 Futures=StreamFutureContracts(kwargs.get("Futures", {})),
#                 Options=StreamOptionContracts(kwargs.get("Options", {})),
#             )
#         )

#     def set_contracts(self, security_type: SecurityType, contracts: ProductContracts):
#         origin_contracts = getattr(self, "{}s".format(security_type.name))
#         origin_contracts.post_init(contracts)

#     def _set_fetched(self):
#         self.Futures._fetched = True
#         self.Options._fetched = True
#         self.Stocks._fetched = True
#         self.Indexs._fetched = True
#         self.status = FetchStatus.Fetched

#     def __iter__(self):
#         return self._iter(to_dict=False, exclude={"status"})

#     def __bool__(self):
#         return self.status == FetchStatus.Fetched

#endregion

class StreamProductContracts:
    security_type: SecurityType

    def __init__(self, **kwargs):
        self._Com:quoteCom = kwargs.get("Com", {})

    def __getitem__(self, key):
        return getattr(self, key, self.__default__(key))

    def __default__(self, key):
        return None


class StreamIndexContracts(StreamProductContracts):
    security_type: SecurityType = SecurityType.Index

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.__TSE = None
        self.__OTC = None

    @property
    def TSE(self):
        if self.__TSE is None:
            self.__TSE = {}
            pkg = self._Com.RetriveLastIndex("TSE")
            for idx in pkg.IDX:                
                self.__TSE.update({idx.code : Index(exchange="TSE", IDX=idx)})

        return self.__TSE
    
    @property
    def OTC(self):
        if self.__OTC is None:
            self.__OTC = {}
            pkg = self._Com.RetriveLastIndex("OTC")
            for idx in pkg.IDX:                
                self.__OTC.update({idx.code : Index(exchange="OTC", IDX=idx)})

        return self.__OTC

class StreamStockContracts(StreamProductContracts):
    security_type: SecurityType = SecurityType.Stock

    def __default__(self, key):
        value = self._Com.GetProductSTOCK(key)
        stock = Stock()
        stock.exchange=value.Market
        stock.code=value.StockNo #(str): 商品代碼
        stock.symbol=value.StockNo #(str): 符號
        stock.name=value.StockName #(str): 商品名稱
            # category (str): 類別
            # unit (int): 單位
        stock.limit_up=value.Bull_Price #(float): 漲停價
        stock.limit_down=value.Bear_Price #(float): 跌停價
        stock.reference=value.Ref_Price #(float): 參考價
            # update_date (str): 更新日期
            # margin_trading_balance (int): 融資餘額
            # short_selling_balance (int): 融券餘額
            # day_trade (DayTrade): 可否當沖 {Yes, No, OnlyBuy}

        setattr(self, key, stock)
        return stock


class StreamFutureContracts(StreamProductContracts):
    security_type: SecurityType = SecurityType.Future

    def __default__(self, key):
        PDetail = self._Com.GetTaifexProductDetail(key)
        PBase = self._Com.GetTaifexProductBase(PDetail.ComId[:3])
        # print(PDetail,'\n')
        # print(PBase)     
        future = Future()
        future.exchange=PBase.ComType
        future.code=PDetail.ComId #(str): 商品代碼
        future.symbol=PDetail.ComId #(str): 符號
        future.name=PBase.ComCName #(str): 商品名稱
        future.category=PBase.ComId #(str): 類別
            # delivery_month (str): 交割月份
            # delivery_date (str): 結算日
            # underlying_kind (str): 標的類型
            # unit (int): 單位
        future.limit_up=PDetail.RisePrice #(float): 漲停價
        future.limit_down=PDetail.FallPrice #(float): 跌停價
        # future.reference=value._REFERENCE_PRICE #(float): 參考價
            # update_date (str): 更新日期
            # margin_trading_balance (int): 融資餘額
            # short_selling_balance (int): 融券餘額
            # day_trade (DayTrade): 可否當沖 {Yes, No, OnlyBuy}        

        setattr(self, key, future)
        return future


class StreamOptionContracts(StreamProductContracts):
    security_type: SecurityType = SecurityType.Option

    def __default__(self, key):
        PDetail = self._Com.GetTaifexProductDetail(key)
        PBase = self._Com.GetTaifexProductBase(PDetail.ComId[:3])
        # print(PDetail,'\n')
        # print(PBase)     
        option = Option()
        option.exchange=PBase.ComType
        option.code=PDetail.ComId #(str): 商品代碼
        option.symbol=PDetail.ComId #(str): 符號
        option.name=PBase.ComCName #(str): 商品名稱
        option.category=PBase.ComId #(str): 類別
            # delivery_month (str): 交割月份
            # delivery_date (str): 結算日
            # underlying_kind (str): 標的類型
            # unit (int): 單位
        option.limit_up=PDetail.RisePrice #(float): 漲停價
        option.limit_down=PDetail.FallPrice #(float): 跌停價
        # future.reference=value._REFERENCE_PRICE #(float): 參考價
            # update_date (str): 更新日期
            # margin_trading_balance (int): 融資餘額
            # short_selling_balance (int): 融券餘額
            # day_trade (DayTrade): 可否當沖 {Yes, No, OnlyBuy}        

        setattr(self, key, option)
        return option


class Contracts:
    Indexs: StreamIndexContracts
    Stocks: StreamStockContracts
    Futures: StreamFutureContracts
    Options: StreamOptionContracts
    status: FetchStatus = FetchStatus.Unfetch

    def __init__(self, **kwargs):
        self.Indexs=StreamIndexContracts(**kwargs)
        self.Stocks=StreamStockContracts(**kwargs)
        self.Futures=StreamFutureContracts(**kwargs)
        self.Options=StreamOptionContracts(**kwargs)


import sys, os
assembly_path = os.path.dirname(__file__)
sys.path.append(assembly_path)
libs_path = os.path.join(assembly_path, '.libs')
sys.path.append(libs_path)

from superpy.superpy import SuperPy
from superpy.constant import Exchange, OrderState, QuoteType
import constant
from superpy.account import Account
from superpy.stream_data_type import (
    TickSTKv1,
    TickFOPv1,
    BidAskSTKv1,
    BidAskFOPv1,
    QuoteSTKv1,
)


from TradeCom import TradeCom as tradeCom
from QuoteCom import Quote as quoteCom
from IntelligencePy import *
from PackagePy import *

from account import Account, AccountType, FutureAccount, StockAccount
from contracts import (
    # BaseContract,
    # ComboContract,
    Contract,
    Contracts,
    FetchStatus,
    Future,
    Index,
    Option,
    Stock,
)
from order import (
    # ComboOrder,
    Order,
    # OrderDealRecords,
    # StrictInt,
    Trade,
    # ComboTrade,
    # conint,
    OrderStatus,
    Deal
)
from constant import (
    Action,
    Exchange,
    OrderState,
    SecurityType,
    Status,
    Unit,
    ScannerType,
    TicksQueryType,
    QuoteType,
    QuoteVersion,
    OrderType,
    StockPriceType,
    StockOrderLot,
    StockOrderCond,
    FuturesPriceType,
    FuturesOCType,
)

from stream_data_type import *

import typing
from datetime import datetime
import time

from quote import Quote
from report import OrderReport

class SuperPy:
    """SuperPy api

    Functions:
        login
        logout
        activate_ca
        list_accounts
        set_default_account
        get_account_margin
        get_account_openposition
        get_account_settle_profitloss
        get_stock_account_funds
        get_stock_account_unreal_profitloss
        get_stock_account_real_profitloss
        place_order
        update_order
        update_status
        list_trades

    Objects:
        Quote
        Contracts
        Order
    """
    def __init__(self, simulation: bool = True, **kwargs):
        self.quote: Quote
        self.stock_account = None
        self.futopt_account = None
        self.Order = Order
        self.trade_host = 'itrade.kgi.com.tw'   #交易正式環境
        self.quote_host = 'iquote.kgi.com.tw'   #行情正式環境
        self.simulation = simulation
        self.__update_status_condition = {}
        self.__setup_com(**kwargs)

    def __setup_com(self, **kwargs):
        if self.simulation: #僅用測試環境domain
            self.trade_host = 'itradetest.kgi.com.tw'
            self.quote_host = 'iquotetest.kgi.com.tw'
        
        #手動修改
        self.trade_host = kwargs.get('trade_host') if kwargs.get('trade_host') else self.trade_host
        self.quote_host = kwargs.get('quote_host') if kwargs.get('quote_host') else self.quote_host

        sid = 'API'
        token = 'b6eb'   
        port = 8000

        self.__tradeCom = tradeCom(self.trade_host, port, sid, callback=lambda dt,dic : {})
        self.__quoteCom = quoteCom(self.quote_host, port, sid, token)
        self.Contracts = Contracts(Com=self.__quoteCom)
        self.quote = Quote(self.__quoteCom)
        self.__orderReport = OrderReport(self.__tradeCom)

    def fetch_contracts(
        self,
        contract_download: bool = False,
        contracts_timeout: int = 0,
        contracts_cb: typing.Callable[[], None] = None,
    ):
        if contract_download and self.Contracts.status == FetchStatus.Unfetch:
            self.Contracts.status = FetchStatus.Fetching

            if self.stock_account is not None:
                for market in ["TSE","OTC"]:
                    self.__quoteCom.RetriveProduct(market)
            if self.futopt_account is not None:
                self.__quoteCom.LoadTaifexProductXML()

            self.Contracts.status = FetchStatus.Fetched

    #region Login

    def login(
        self,
        userID: str,
        password: str,
        fetch_contract: bool = True,
        contracts_timeout: int = 0,
        contracts_cb: typing.Callable[[], None] = None,
        subscribe_trade: bool = True,
        receive_window: int = 30000,
        loginInfos = None,
    ) -> typing.List[Account]:
        
        self.__uid = userID
        self.__tradeCom.Login(userID, password, subscribe_trade, loginInfos=loginInfos)
        self.__quoteCom.Login(userID, password)
        accounts = self.list_accounts()

        if fetch_contract:
            self.fetch_contracts(fetch_contract, contracts_timeout, contracts_cb)

        return accounts

    def logout(self) -> bool:
        self.__tradeCom.Logout()
        self.__tradeCom.Dispose()

        self.__quoteCom.Logout()
        self.__quoteCom.Dispose()
    
    #endregion

    #region Account

    def list_accounts(self) -> typing.List[Account]:
        """list all account you have"""
        accounts = []
        for acc in self.__tradeCom.accounts:
            _acc = (self.__uid, acc.BrokeId, acc.Account, bool(acc.ISCASIGN), acc.TRADER)
            if AccountType(acc.AccountFlag) == AccountType.Stock:
                _account = StockAccount(*_acc)
            elif AccountType(acc.AccountFlag) == AccountType.Future:
                _account = FutureAccount(*_acc)
            else:
                pass
            self.set_default_account(_account)
            accounts.append(_account)
        
        return accounts
    
    def set_default_account(self, account):
        """set default account for trade when place order not specific

        Args:
            account (:obj:Account):
                choice the account from listing account and set as default
        """
        if isinstance(account, StockAccount):
            self.stock_account = account
        elif isinstance(account, FutureAccount):
            self.futopt_account = account

    #endregion

    #region Order

    def place_order(
        self,
        contract: Contract,
        order: Order,
        timeout: int = 5000,
        cb: typing.Callable[[Trade], None] = None,
    ) -> Trade:
        
        if not order.account:
            if isinstance(contract, Future) or isinstance(contract, Option):
                order.account = self.futopt_account
            elif isinstance(contract, Stock):
                order.account = self.stock_account
            else:
                # log.error("Please provide the account place to.")
                return None
        
        if isinstance(contract, Stock):
            #region 證券下單
            SECU_ORDER_RPT = self.__tradeCom.Order(
                ordtype=Security_OrdType.OT_NEW,
                ordlot=order.order_lot.MapTo(),
                ordclass=order.order_cond.MapTo() if not order.daytrade_short else Security_Class.SC_DayTrade,
                BrokerId=order.account.broker_id, 
                Account=order.account.account_id, 
                stockid=contract.code, 
                bs= order.action.MapTo(),
                qty=order.quantity, 
                Price=order.price,
                PF=order.price_type.MapTo(),
                subAccount="", 
                AgentID="", 
                OrderNo="", 
                TIF=order.order_type.MapTo()
                )
            # print("Order: {Log}".format(Log=SECU_ORDER_RPT))
            if not SECU_ORDER_RPT:
                return None
            order.ordno=SECU_ORDER_RPT.OrderNo
            order.id=SECU_ORDER_RPT.CNTN

            status=OrderStatus()
            status.id = order.id
            status.status = Status.Submitted if SECU_ORDER_RPT.ErrCode == '0' else Status.Failed
            status.status_code = SECU_ORDER_RPT.ErrCode
            if status.status == Status.Failed:
                status.ErrMsg = SECU_ORDER_RPT.ErrMsg
            status.order_datetime = datetime.strptime(SECU_ORDER_RPT.ClientOrderTimeN, "%Y%m%d%H%M%S%f")
            status.order_quantity = SECU_ORDER_RPT.Qty

            trade = Trade(contract=contract,order=order,status=status)
            #endregion
        elif isinstance(contract, Future) or isinstance(contract, Option):
            #region 期貨下單
            FUT_ORDER_RPT = self.__tradeCom.FutOrder(
                type=ORDER_TYPE.OT_NEW, 
                market=MARKET_FLAG.MF_FUT if isinstance(contract, Future) else MARKET_FLAG.MF_OPT,
                brokerId=order.account.broker_id, 
                account=order.account.account_id, 
                subAccount=order.account.trader,
                symbolId=contract.code, 
                BS=order.action.MapTo(),
                pricefl=order.price_type.MapTo(),
                price=order.price,
                TIF=order.order_type.MapTo(),
                qty=order.quantity,
                pf=order.octype.MapTo(), 
                off=OFFICE_FLAG.OF_AS400
            )            
            # print("Order: {Log}".format(Log=FUT_ORDER_RPT))
            if not FUT_ORDER_RPT:
                return None
            order.ordno=FUT_ORDER_RPT.OrderNo
            order.id=FUT_ORDER_RPT.CNT

            status=OrderStatus()
            status.id = order.id
            status.status = Status.Submitted if FUT_ORDER_RPT.Code == '0' else Status.Failed
            status.status_code = FUT_ORDER_RPT.Code
            if status.status == Status.Failed:
                status.ErrMsg = FUT_ORDER_RPT.ErrMsg
            status.order_datetime = datetime.strptime(FUT_ORDER_RPT.TradeDate + FUT_ORDER_RPT.ReportTime, "%Y%m%d%H%M%S%f")
            status.order_quantity = FUT_ORDER_RPT.AfterQty
            status.web_id = FUT_ORDER_RPT.WebID

            trade = Trade(contract=contract,order=order,status=status)
            #endregion
        return trade

    def update_order(
        self,
        trade: Trade,
        price: float = None,
        qty: int = None,
        timeout: int = 5000,
        cb: typing.Callable[[Trade], None] = None,
    ) -> Trade:
        order=trade.order
        contract=trade.contract
        if price is not None:
            ordtype=Security_OrdType.OT_MODIFY_PRICE
            fut_ordtype=ORDER_TYPE.OT_MODIFY_PRICE
            qty = order.quantity
        elif qty is not None:
            ordtype=Security_OrdType.OT_MODIFY_QTY
            fut_ordtype=ORDER_TYPE.OT_MODIFY_QTY
            price = order.price
        else:
            return None
        
        if isinstance(contract, Stock):
            #region 證券改單
            SECU_ORDER_RPT = self.__tradeCom.Order(
                ordtype=ordtype,
                ordlot=order.order_lot.MapTo(),
                ordclass=order.order_cond.MapTo(),
                BrokerId=order.account.broker_id, 
                Account=order.account.account_id, 
                stockid=contract.code, 
                bs= order.action.MapTo(),
                qty=qty, 
                Price=price,
                PF=order.price_type.MapTo(),
                subAccount="", 
                AgentID="", 
                OrderNo=order.ordno, 
                TIF=order.order_type.MapTo()
                )
            # print("Modify_Order: {Log}".format(Log=SECU_ORDER_RPT.__dict__))

            trade.status.status = Status.Submitted if SECU_ORDER_RPT.ErrCode == '0' else Status.Failed
            trade.status.status_code = SECU_ORDER_RPT.ErrCode
            trade.status.order_datetime = datetime.strptime(SECU_ORDER_RPT.ClientOrderTimeN, "%Y%m%d%H%M%S%f")
            if ordtype == Security_OrdType.OT_MODIFY_PRICE:
                trade.status.modified_price = price
            if ordtype == Security_OrdType.OT_MODIFY_QTY:
                trade.status.cancel_quantity = qty
            #endregion
        elif isinstance(contract, Future) or isinstance(contract, Option):
            #region 期貨改單
            FUT_ORDER_RPT = self.__tradeCom.FutOrder(
                type=fut_ordtype, 
                market=MARKET_FLAG.MF_FUT if isinstance(contract, Future) else MARKET_FLAG.MF_OPT,
                brokerId=order.account.broker_id, 
                account=order.account.account_id, 
                subAccount=order.account.trader,
                symbolId=contract.code, 
                BS=order.action.MapTo(),
                pricefl=order.price_type.MapTo(),
                price=price,
                TIF=order.order_type.MapTo(),
                qty=qty,
                pf=order.octype.MapTo(), 
                off=OFFICE_FLAG.OF_AS400,
                webid=trade.status.web_id,
                cnt=order.id, 
                orderno=order.ordno
            )            
            # print("Order: {Log}".format(Log=FUT_ORDER_RPT))

            trade.status.status = Status.Submitted if FUT_ORDER_RPT.Code == '0' else Status.Failed
            trade.status.status_code = FUT_ORDER_RPT.Code
            trade.status.order_datetime = datetime.strptime(FUT_ORDER_RPT.TradeDate + FUT_ORDER_RPT.ReportTime, "%Y%m%d%H%M%S%f")
            if fut_ordtype == ORDER_TYPE.OT_MODIFY_PRICE:
                trade.status.modified_price = price
            if fut_ordtype == ORDER_TYPE.OT_MODIFY_QTY:
                trade.status.cancel_quantity = qty
            #endregion

        return trade

    def cancel_order(
        self,
        trade: Trade,
        timeout: int = 5000,
        cb: typing.Callable[[Trade], None] = None,
    ) -> Trade:
        order=trade.order
        contract=trade.contract
        if isinstance(contract, Stock):
            #region 證券刪單
            SECU_ORDER_RPT = self.__tradeCom.Order(
                ordtype=Security_OrdType.OT_CANCEL,
                ordlot=order.order_lot.MapTo(),
                ordclass=order.order_cond.MapTo(),
                BrokerId=order.account.broker_id, 
                Account=order.account.account_id, 
                stockid=contract.code, 
                bs= order.action.MapTo(),
                qty=order.quantity, 
                Price=order.price,
                PF=order.price_type.MapTo(),
                subAccount="", 
                AgentID="", 
                OrderNo=order.ordno, 
                TIF=order.order_type.MapTo()
                )
            # print("Cancel_Order: {Log}".format(Log=SECU_ORDER_RPT.__dict__))

            trade.status.status_code = SECU_ORDER_RPT.ErrCode
            trade.status.order_datetime = datetime.strptime(SECU_ORDER_RPT.ClientOrderTimeN, "%Y%m%d%H%M%S%f")
            trade.status.status = Status.Cancelled if SECU_ORDER_RPT.ErrCode == '0' else Status.Failed
            trade.status.cancel_quantity = SECU_ORDER_RPT.BeforeQtyN
            #endregion
        elif isinstance(contract, Future) or isinstance(contract, Option):
            #region 期貨刪單
            FUT_ORDER_RPT = self.__tradeCom.FutOrder(
                type=ORDER_TYPE.OT_CANCEL, 
                market=MARKET_FLAG.MF_FUT if isinstance(contract, Future) else MARKET_FLAG.MF_OPT,
                brokerId=order.account.broker_id, 
                account=order.account.account_id, 
                subAccount=order.account.trader,
                symbolId=contract.code, 
                BS=order.action.MapTo(),
                pricefl=order.price_type.MapTo(),
                price=order.price,
                TIF=order.order_type.MapTo(),
                qty=order.quantity,
                pf=order.octype.MapTo(), 
                off=OFFICE_FLAG.OF_AS400,
                webid=trade.status.web_id,
                cnt=order.id, 
                orderno=order.ordno
            )            
            # print("Order: {Log}".format(Log=FUT_ORDER_RPT))

            trade.status.status_code = FUT_ORDER_RPT.Code
            trade.status.order_datetime = datetime.strptime(FUT_ORDER_RPT.TradeDate + FUT_ORDER_RPT.ReportTime, "%Y%m%d%H%M%S%f")
            trade.status.status = Status.Cancelled if FUT_ORDER_RPT.Code == '0' else Status.Failed
            trade.status.cancel_quantity = FUT_ORDER_RPT.BeforeQty
            #endregion

        return trade

    def update_status(
        self,
        account: Account = None,
        trade: Trade = None,
        orderno: str = None,
    ):
        """update status of all trades you have"""

        if trade:
            self.__update_status_condition = { 'trade' : trade }
        elif account:
            if account.signed or self.simulation:
                self.__update_status_condition = { 'account' : account }
        elif orderno:
            self.__update_status_condition = { 'orderno' : orderno }
        else:
            if self.stock_account:
                if self.stock_account.signed or self.simulation:
                    self.__update_status_condition = { 'account' : self.stock_account }
            if self.futopt_account:
                if self.futopt_account.signed or self.simulation:
                    self.__update_status_condition = { 'account' : self.futopt_account }

    def list_trades(self) -> typing.List[Trade]:
        """list all trades"""
        trade = self.__update_status_condition.get('trade')
        orderNo = trade.order.ordno if trade else self.__update_status_condition.get('orderno')
        account = self.__update_status_condition.get('account')

        trades = [self.__combine_trade_by_orderNo(ordNo) \
                  for ordNo in self.__tradeCom.ORDER_RPT_List.keys() \
                    if orderNo is None or orderNo == ordNo]

        if account:
            trades = [trade for trade in trades if vars(trade.order.account) == vars(account)]

        return trades

    def __combine_trade_by_orderNo(self, orderNo) -> Trade:
        trade: Trade = None
        rpt_list = self.__tradeCom.ORDER_RPT_List.get(orderNo, [])
        is_future_report = False
        for rpt in rpt_list:
            if rpt.OrderFunc == 'I':
                trade = self.__build_trade_by_report(rpt)
                is_future_report = (DT(rpt.DT) == DT.FUT_ORDER_RPT)

        if trade is None:
            raise ValueError("order report not found.")        

        if is_future_report:
            rpt_list = sorted(rpt_list, key=lambda rpt: rpt.ReportTime)
        else:
            rpt_list = sorted(rpt_list, key=lambda rpt: rpt.ReportTimeN)

        for rpt in rpt_list:
            if rpt.OrderFunc == 'I':
                continue
            else:
                status = Status.Cancelled if rpt.OrderFunc == 'D' else Status.Submitted
                if isinstance(trade.contract, Stock):
                    trade.status.status = status if rpt.ErrCode == '0' else Status.Failed
                    trade.status.status_code = rpt.ErrCode
                    trade.status.order_datetime = datetime.strptime(rpt.ClientOrderTimeN, "%Y%m%d%H%M%S%f")
                elif isinstance(trade.contract, Future) or isinstance(trade.contract, Option):
                    trade.status.status = status if rpt.Code == '0' else Status.Failed
                    trade.status.status_code = rpt.Code
                    trade.status.order_datetime = datetime.strptime(rpt.TradeDate + rpt.ReportTime, "%Y%m%d%H%M%S%f")

                if rpt.OrderFunc in ['M','R']:
                    trade.status.modified_price = rpt.Price
                else:
                    if isinstance(trade.contract, Stock):
                        trade.status.cancel_quantity += int(rpt.Qty)
                    elif isinstance(trade.contract, Future) or isinstance(trade.contract, Option):
                        trade.status.cancel_quantity += int(rpt.BeforeQty) - int(rpt.AfterQty)

        deal_list = self.__tradeCom.DEAL_RPT_List.get(orderNo, [])
        
        for rpt in deal_list:
            deal = Deal()
            deal.seq = rpt.MarketNo
            if is_future_report:
                deal.price = rpt.DealPrice
                deal.quantity = rpt.DealQty                
            else:
                deal.price = rpt.PriceN
                deal.quantity = rpt.DealQtyN
                
            trade.status.status = Status.Filled if trade.status.order_quantity == deal.quantity else Status.PartFilled                
            if trade.status.deals is None:
                trade.status.deals = []
            trade.status.deals.append(deal)

        return trade

    def __build_trade_by_report(self, orderRpt) -> Trade:
        match DT(orderRpt.DT):
            case DT.SECU_ORDER_RPT:
                contract = self.Contracts.Stocks[orderRpt.StockID]
                orderRpt: PT04010 = orderRpt
            case DT.FUT_ORDER_RPT:
                contract = self.Contracts.Futures[orderRpt.Symbol]
                if contract is None or contract.exchange == 'O':
                    contract = self.Contracts.Options[orderRpt.Symbol]                
                orderRpt: PT02010 = orderRpt

        order_account = [acc for acc in self.list_accounts() \
                         if acc.broker_id == orderRpt.BrokerId and \
                            acc.account_id == orderRpt.Account].pop()

        if isinstance(contract, Stock):
            order = self.Order(
                price=orderRpt.Price,
                quantity=orderRpt.Qty,
                action={'B' : Action.Buy , 'S' : Action.Sell}.get(orderRpt.Side, None),
                price_type={'1' : StockPriceType.MKT , '2' : StockPriceType.LMT , '0' : 'MKP'}.get(orderRpt.PriceFlagN, None),
                order_type={'R':OrderType.ROD, 'I':OrderType.IOC, 'F':OrderType.FOK}.get(orderRpt.TimeInForceN, None),
                order_lot={'0' : StockOrderLot.Common, '2' : StockOrderLot.Fixing, '1' : StockOrderLot.Odd, '4' : StockOrderLot.IntradayOdd}.get(orderRpt.OrdLot, None),
                order_cond={'0' : StockOrderCond.Cash, '3' : StockOrderCond.MarginTrading, '4' : StockOrderCond.ShortSelling}.get(orderRpt.OrdClass, None),
                daytrade_short=True if orderRpt.OrdClass == '9' else False,
                account=order_account
            )
            order.id=orderRpt.CNTN
            order.ordno=orderRpt.OrderNo

            status=OrderStatus()
            status.id = order.id
            status.status = Status.Submitted if orderRpt.ErrCode == '0' else Status.Failed
            status.status_code = orderRpt.ErrCode
            status.order_datetime = datetime.strptime(orderRpt.ClientOrderTimeN, "%Y%m%d%H%M%S%f")
            status.order_quantity = orderRpt.Qty
        elif isinstance(contract, Future) or isinstance(contract, Option):
            order = self.Order(
                price=orderRpt.Price,
                quantity=orderRpt.AfterQty,
                action={'B' : Action.Buy , 'S' : Action.Sell}.get(orderRpt.Side, None),
                price_type={'1' : FuturesPriceType.MKT , '2' : FuturesPriceType.LMT , '0' : FuturesPriceType.MKP}.get(orderRpt.PriceMark, None),
                order_type={'R':OrderType.ROD, 'I':OrderType.IOC, 'F':OrderType.FOK}.get(orderRpt.TimeInForce, None),                
                octype={
                    'A' : FuturesOCType.Auto, 
                    'O' : FuturesOCType.New, 
                    'C' : FuturesOCType.Cover,
                    'T' : FuturesOCType.DayTrade
                    }.get(orderRpt.PositionEffect, None),
                daytrade_short=True if orderRpt.PositionEffect == 'T' else False,
                account=order_account
            )
            order.id=orderRpt.CNT
            order.ordno=orderRpt.OrderNo
            
            status=OrderStatus()
            status.id = order.id
            status.status = Status.Submitted if orderRpt.Code == '0' else Status.Failed
            status.status_code = orderRpt.Code
            status.order_datetime = datetime.strptime(orderRpt.TradeDate + orderRpt.ReportTime, "%Y%m%d%H%M%S%f")
            status.order_quantity = orderRpt.AfterQty
            status.web_id = orderRpt.WebID

        trade = Trade(contract=contract,order=order,status=status)
        return trade

    def set_order_callback(
        self, func: typing.Callable[[OrderState, dict], None]
    ) -> None:
        self.__orderReport.set_order_callback(func)

    #endregion

    #region Quote

    def on_tick_stk_v1(
        self, bind: bool = False
    ) -> typing.Callable[[Exchange, TickSTKv1], None]:
        def wrap_deco(
            func: typing.Callable[[Exchange, TickSTKv1], None]
        ) -> typing.Callable[[Exchange, TickSTKv1], None]:
            self.quote.set_on_tick_stk_v1_callback(func, bind)
            return func

        return wrap_deco

    def on_tick_fop_v1(
        self, bind: bool = False
    ) -> typing.Callable[[Exchange, TickFOPv1], None]:
        def wrap_deco(
            func: typing.Callable[[Exchange, TickFOPv1], None]
        ) -> typing.Callable[[Exchange, TickFOPv1], None]:
            self.quote.set_on_tick_fop_v1_callback(func, bind)
            return func

        return wrap_deco

    def on_bidask_stk_v1(
        self, bind: bool = False
    ) -> typing.Callable[[Exchange, BidAskSTKv1], None]:
        def wrap_deco(
            func: typing.Callable[[Exchange, BidAskSTKv1], None]
        ) -> typing.Callable[[Exchange, BidAskSTKv1], None]:
            self.quote.set_on_bidask_stk_v1_callback(func, bind)
            return func

        return wrap_deco

    def on_bidask_fop_v1(
        self, bind: bool = False
    ) -> typing.Callable[[Exchange, BidAskFOPv1], None]:
        def wrap_deco(
            func: typing.Callable[[Exchange, BidAskFOPv1], None]
        ) -> typing.Callable[[Exchange, BidAskFOPv1], None]:
            self.quote.set_on_bidask_fop_v1_callback(func, bind)
            return func

        return wrap_deco

    #endregion



from IntelligencePy import DT 
from IntelligencePy import COM_STATUS
from PackagePy import *
from queue import Queue, Empty
import datetime

class Login:
    def __init__(self, host, port):
        self.quoteCom = None
        self.host = host
        self.port = port

        self.__com_status = None
        self.__Com_StatusDict = dict()
        self.__GetStatusQueue = Queue()
        self.__RcvMessageQueue = Queue()

    def Dispose(self):
        """關閉API的元件
        """
        self.quoteCom.Dispose()

    def Logout(self):
        """登出API平台
        """
        self.quoteCom.Logout()

    def Login(self, uid, pwd):
        """自行登入
        Args:
            uid (_type_): _description_
            pwd (_type_): _description_
        """
        self.quoteCom.Connect2Quote(self.host, self.port, uid, pwd, ' ', '')
        accounts = []
        try:
            status = self.__GetStatusQueue.get(timeout=15)
            # print(status, self.Com_StatusDict[status])  
            match status:
                case COM_STATUS.LOGIN_READY:
                    pkg = self.__RcvMessageQueue.get(timeout=5)
                    # print(datetime.datetime.now(), "DT=[{dt}] {Log}".format(dt=pkg.DT,Log=pkg.ToLog())) 
                    P1503 = P001503(pkg)
                    if (P1503.Code==0):
                        print("QuoteCom登入成功")
                        accounts = P1503.p001503_2
                case _:
                    print('QuoteCom登入失敗' , status, self.__Com_StatusDict.get(status))
        except Empty:
            print('QuoteCom登入失敗 Timeout' ,self.__com_status, self.__Com_StatusDict.get(self.__com_status))
        
        return accounts

    def Login_RcvMessage(self, dt:DT, pkg):
        if dt == DT.LOGIN:
            self.__RcvMessageQueue.put(pkg)
            return True
        else:
            return False
        
    def Login_GetStatus(self, status:COM_STATUS, msg):
        if status in [COM_STATUS.LOGIN_READY, COM_STATUS.CONNECT_FAIL, COM_STATUS.NOVALIDCERT, COM_STATUS.LOGIN_FAIL, COM_STATUS.DISCONNECTED]:
            self.__com_status = status
            self.__Com_StatusDict[self.__com_status] = { 'Time' : datetime.datetime.now() , 'Msg' : msg }
            self.__GetStatusQueue.put(status)


import typing, abc

from QuoteCom import Quote as quoteCom
from PackagePy import *

from contracts import (
    Contract,
    Contracts,
    Future,
    Index,
    Option,
    Stock,
)
from constant import *
from stream_data_type import (
    TickSTKv1,
    TickFOPv1,
    BidAskSTKv1,
    BidAskFOPv1,
    QuoteSTKv1,
    NewIndex,
)

class Quote:
    def __init__(self, QuoteCom) -> None:
        self.__quoteCom : quoteCom = QuoteCom
        self.__quoteCom.callback = self.__quoteCom_callback

        self.__on_tick_stk_v1_callback = None
        self.__tick_idx = {
            DT.QUOTE_STOCK_INDEX1 : [],
            DT.QUOTE_STOCK_INDEX2 : []
        }    #暫存訂閱Tick的指數商品代號
        self.__on_bidask_stk_v1_callback = None

        self.__on_tick_fop_v1_callback = None
        self.__tick_fop = []    #暫存訂閱Tick的期權商品代號
        self.__on_bidask_fop_v1_callback = None
        self.__bidask_fop = []  #暫存訂閱BidAsk的期權商品代號

        self.__on_new_index_callback = None
        self.__new_idx = [] #暫存訂閱新編指數商品代號

    def __quoteCom_callback(self, pkgDT:DT, pkg):
        # 證券Tick
        if self.__on_tick_stk_v1_callback is not None:
            if pkgDT==DT.QUOTE_STOCK_MATCH1 or pkgDT==DT.QUOTE_STOCK_MATCH2 or \
                pkgDT==DT.QUOTE_ODD_MATCH1 or pkgDT==DT.QUOTE_ODD_MATCH2:
                    Tick = TickSTKv1()
                    pkg = PI31001(pkg)
                    Tick.code = pkg.StockNo
                    Tick.datetime = pkg.Match_Time
                    Tick.amount = pkg.Match_Price
                    Tick.volume = pkg.Match_Qty
                    Tick.total_volume = pkg.Total_Qty
                    Tick.simtrade = not bool(pkg.Status)    #Status 0=試撮 1=非試撮
                    Tick.intraday_odd = 1 if pkgDT==DT.QUOTE_ODD_MATCH1 or pkgDT==DT.QUOTE_ODD_MATCH2 else 0
                    exchange = Exchange.OTC if pkgDT==DT.QUOTE_STOCK_MATCH2 or pkgDT==DT.QUOTE_ODD_MATCH2 else Exchange.TSE
                    self.__on_tick_stk_v1_callback(exchange, Tick)
                    return
            
        # 證券BidAsk
        if self.__on_bidask_stk_v1_callback is not None:
            if pkgDT==DT.QUOTE_STOCK_DEPTH1 or pkgDT==DT.QUOTE_STOCK_DEPTH2 or \
                pkgDT==DT.QUOTE_ODD_DEPTH1 or pkgDT==DT.QUOTE_ODD_DEPTH2:
                    BidAsk = BidAskSTKv1()
                    pkg = PI31002(pkg)
                    # print(pkg)
                    BidAsk.code = pkg.StockNo
                    BidAsk.datetime = pkg.Match_Time
                    BidAsk.bid_price = [level.PRICE for level in pkg.BUY_DEPTH] # 委買價
                    BidAsk.bid_volume = [level.QUANTITY for level in pkg.BUY_DEPTH] # 委買量 (張)
                    # diff_bid_vol (:List:int): 買價增減量 (張)
                    BidAsk.ask_price = [level.PRICE for level in pkg.SELL_DEPTH] # 委賣價
                    BidAsk.ask_volume = [level.QUANTITY for level in pkg.SELL_DEPTH]# 委賣量
                    # diff_ask_vol (:List:int): 賣價增減量 (張)
                    # suspend (bool): 暫停交易
                    BidAsk.simtrade = not bool(pkg.Status)    #Status 0=試撮 1=非試撮
                    BidAsk.intraday_odd = 1 if pkgDT==DT.QUOTE_ODD_DEPTH1 or pkgDT==DT.QUOTE_ODD_DEPTH2 else 0
                    exchange = Exchange.OTC if pkgDT==DT.QUOTE_ODD_DEPTH1 or pkgDT==DT.QUOTE_ODD_DEPTH2 else Exchange.TSE
                    self.__on_bidask_stk_v1_callback(Exchange.TSE, BidAsk)
                    return
            
        # 期貨Tick
        if self.__on_tick_fop_v1_callback is not None:
             if pkgDT in [DT.QUOTE_I020, DT.QUOTE_I022]:
                if pkg.Symbol not in self.__tick_fop:   
                    return
                pkg = PI20020(pkg)
                Tick = TickFOPv1()
                Tick.code = pkg.Symbol
                Tick.datetime = pkg.MatchTime
                Tick.amount = pkg.Price
                Tick.volume = pkg.MatchQuantity
                Tick.total_volume = pkg.MatchTotalQty
                Tick.bid_side_total_vol = pkg.MatchBuyCnt
                Tick.ask_side_total_vol = pkg.MatchSellCnt
                Tick.simtrade = 1 if pkgDT == DT.QUOTE_I022 else 0  #試撮
                
                self.__on_tick_fop_v1_callback(Exchange.TAIFEX, Tick)
                return
        
        # 期貨BidAsk
        if self.__on_bidask_fop_v1_callback is not None:
            if pkgDT in [DT.QUOTE_I080, DT.QUOTE_I082]: 
                if pkg.Symbol not in self.__bidask_fop: 
                    return
                pkg = PI20080(pkg)
                BidAsk = BidAskFOPv1()
                BidAsk.code = pkg.Symbol
                BidAsk.datetime = pkg.DATA_TIME
                BidAsk.bid_price = [level.PRICE for level in pkg.BUY_DEPTH] # 委買價
                BidAsk.bid_volume = [level.QUANTITY for level in pkg.BUY_DEPTH] # 委買量
                BidAsk.ask_price = [level.PRICE for level in pkg.SELL_DEPTH] # 委賣價
                BidAsk.ask_volume = [level.QUANTITY for level in pkg.SELL_DEPTH] # 委賣量
                BidAsk.first_derived_bid_price = pkg.FIRST_DERIVED_BUY_PRICE # 衍生一檔委買價
                BidAsk.first_derived_ask_price = pkg.FIRST_DERIVED_SELL_PRICE # 衍生一檔委賣價
                BidAsk.first_derived_bid_vol = pkg.FIRST_DERIVED_BUY_QTY # 衍生一檔委買量
                BidAsk.first_derived_ask_vol = pkg.FIRST_DERIVED_SELL_QTY # 衍生一檔委賣量
                BidAsk.simtrade = 1 if pkgDT == DT.QUOTE_I082 else 0  #試撮
                
                self.__on_bidask_fop_v1_callback(Exchange.TAIFEX, BidAsk)
                return

        # 指數
        if self.__on_tick_stk_v1_callback is not None:
            if pkgDT in [DT.QUOTE_STOCK_INDEX1, DT.QUOTE_STOCK_INDEX2]:
                _PI31011 = PI31011(pkg)
                PI31011_IDX = _PI31011.IDX
                for contract in self.__tick_idx[pkgDT]:
                    index = int(contract.code) - 1
                    # print(_PI31011.__dict__, _PI31011.Match_Time, PI31011_IDX[index].VALUE)
                    Tick = TickSTKv1()
                    Tick.code = contract.code
                    Tick.datetime = _PI31011.Match_Time
                    Tick.amount = PI31011_IDX[index].VALUE
                    Tick.volume = PI31011_IDX[index].VALUE
                    Tick.total_volume = PI31011_IDX[index].VALUE
                    Tick.symbol = contract.symbol
                    Tick.name = contract.name
                    exchange = Exchange.TSE if pkgDT==DT.QUOTE_STOCK_INDEX1 else Exchange.OTC
                    
                    self.__on_tick_stk_v1_callback(exchange, Tick)
                return

        # 新編指數
        if self.__on_new_index_callback is not None:
            if pkgDT in [DT.QUOTE_STOCK_NEWINDEX1, DT.QUOTE_STOCK_NEWINDEX2]:
                
                if pkg.IndexNo not in self.__new_idx : return

                exchange = Exchange.TSE if pkgDT==DT.QUOTE_STOCK_NEWINDEX1 else Exchange.OTC
                newIndex = NewIndex()
                
                newIndex.IndexNo = pkg.IndexNo
                newIndex.IndexTime = pkg.IndexTime
                newIndex.LastestIndex = pkg.LatestIndex
                
                self.__on_new_index_callback(exchange, newIndex)
                return

    def subNewIndex(
        self,
        code
    ):        
        if len(self.__new_idx) == 0: 
            self.__quoteCom.quoteCom.SubQuotesNewIndex()
        if code not in self.__new_idx : 
            self.__new_idx.append(code)
        

    def unsubNewIndex(
        self,
        code
    ):
        if code in self.__new_idx : 
            self.__new_idx.remove(code)
        if len(self.__new_idx) == 0: 
            self.__quoteCom.quoteCom.UnSubQuotesNewIndex()

    # @abc.abstractmethod
    def subscribe(
        self,
        contract: Contract,
        quote_type: QuoteType = QuoteType.Tick,
        intraday_odd: bool = False,
        # version: QuoteVersion = QuoteVersion.v1,
    ):
        if isinstance(contract, Future) or isinstance(contract, Option):
            self.__quoteCom.quoteCom.SubQuote(contract.code)    # 訂閱報價
            match(quote_type):
                case QuoteType.Tick:
                    if contract.code not in self.__tick_fop : self.__tick_fop.append(contract.code)
                case QuoteType.BidAsk:                    
                    if contract.code not in self.__bidask_fop : self.__bidask_fop.append(contract.code)            
            return
        
        if isinstance(contract, Index):
            self.__quoteCom.quoteCom.SubQuotesIndex()
            if contract.exchange == 'TSE':
                exchange = DT.QUOTE_STOCK_INDEX1
            elif contract.exchange == 'OTC':
                exchange = DT.QUOTE_STOCK_INDEX2
            match(quote_type):
                case QuoteType.Tick:
                    if contract not in self.__tick_idx[exchange] : self.__tick_idx[exchange].append(contract)
            return
        
        if not intraday_odd:
            match(quote_type):
                case QuoteType.Tick:
                    self.__quoteCom.quoteCom.SubQuotesMatch(contract.code) ##訂閱成交資料
                case QuoteType.BidAsk:
                    self.__quoteCom.quoteCom.SubQuotesDepth(contract.code) ##訂閱五檔資料
                case QuoteType.Quote:
                    self.__quoteCom.quoteCom.RetriveLastPriceStock(contract.code) ##查詢整股最新行情資料
        else :
            match(quote_type):
                case QuoteType.Tick:
                    self.__quoteCom.quoteCom.SubQuotesMatchOdd(contract.code) ##訂閱盤中零股成交資料
                case QuoteType.BidAsk:
                    self.__quoteCom.quoteCom.SubQuotesDepthOdd(contract.code)  ##訂閱盤中零股五檔資料
                case QuoteType.Quote:
                    self.__quoteCom.quoteCom.RetriveLastPriceStockOdd(contract.code) ##查詢盤中零股最新行情資料

    # @abc.abstractmethod
    def unsubscribe(
        self,
        contract: Contract,
        quote_type: QuoteType = QuoteType.Tick,
        intraday_odd: bool = False,
        # version: QuoteVersion = QuoteVersion.v1,
    ):
        if isinstance(contract, Future) or isinstance(contract, Option):            
            match(quote_type):
                case QuoteType.Tick:
                    if contract.code in self.__tick_fop : self.__tick_fop.remove(contract.code)
                case QuoteType.BidAsk:                    
                    if contract.code in self.__bidask_fop : self.__bidask_fop.remove(contract.code)
            
            # QuoteCom解訂閱會Tick, BidAsk一起解，先檢查都不使用在解除
            if contract.code not in self.__tick_fop and contract.code not in self.__bidask_fop:            
                self.__quoteCom.quoteCom.UnsubQuotes(contract.code)    # 解除訂閱報價

            return
        
        if isinstance(contract, Index):
            if contract.exchange == 'TSE':
                exchange = DT.QUOTE_STOCK_INDEX1
            elif contract.exchange == 'OTC':
                exchange = DT.QUOTE_STOCK_INDEX2
            match(quote_type):
                case QuoteType.Tick:
                    if contract in self.__tick_idx[exchange] : self.__tick_idx[exchange].remove(contract)

            if len(self.__tick_idx[DT.QUOTE_STOCK_INDEX1]) == 0 and \
                len(self.__tick_idx[DT.QUOTE_STOCK_INDEX2]) == 0:
                #無任一訂閱指數 則解除訂閱
                self.__quoteCom.quoteCom.UnSubQuotesIndex()# 解除訂閱指數
            return

        if not intraday_odd:
            match(quote_type):
                case QuoteType.Tick:
                    self.__quoteCom.quoteCom.UnSubQuotesMatch(contract.code) ##解除訂閱成交資料
                case QuoteType.BidAsk:                    
                    self.__quoteCom.quoteCom.UnSubQuotesDepth(contract.code) ##解除訂閱五檔資料 
                # case QuoteType.Quote:
                    # self.__quoteCom.quoteCom.RetriveLastPriceStock(contract.code) ##查詢整股最新行情資料
        else :
            match(quote_type):
                case QuoteType.Tick:
                    self.__quoteCom.quoteCom.UnSubQuotesMatchOdd(contract.code) ##解除訂閱盤中零股成交資料
                case QuoteType.BidAsk:                    
                    self.__quoteCom.quoteCom.UnSubQuotesDepthOdd(contract.code)  ##解除訂閱盤中零股五檔資料
                # case QuoteType.Quote:
                    # self.__quoteCom.quoteCom.RetriveLastPriceStockOdd(contract.code) ##查詢盤中零股最新行情資料
        
        print('unsubscribe', contract.code, quote_type)

    # @abc.abstractmethod
    def set_on_tick_stk_v1_callback(
        self,
        func: typing.Callable[[Exchange, TickSTKv1], None],
        bind: bool = False,
    ) -> None:
        self.__on_tick_stk_v1_callback = func

    # @abc.abstractmethod
    def set_on_tick_fop_v1_callback(
        self,
        func: typing.Callable[[Exchange, TickFOPv1], None],
        bind: bool = False,
    ) -> None:
        self.__on_tick_fop_v1_callback = func

    # @abc.abstractmethod
    def set_on_bidask_stk_v1_callback(
        self,
        func: typing.Callable[[Exchange, BidAskSTKv1], None],
        bind: bool = False,
    ) -> None:
        self.__on_bidask_stk_v1_callback = func

    # @abc.abstractmethod
    def set_on_bidask_fop_v1_callback(
        self,
        func: typing.Callable[[Exchange, BidAskFOPv1], None],
        bind: bool = False,
    ) -> None:
        self.__on_bidask_fop_v1_callback = func

    def set_on_new_index_callback(
        self,
        func: typing.Callable[[Exchange, NewIndex], None],
        bind: bool = False,
    ) -> None:
        self.__on_new_index_callback = func


from enum import Enum

# from shioaji.base import BaseModel

__all__ = ("Account StockAccount FutureAccount").split()


class AccountType(str, Enum):
    Stock = "S"
    Future = "F"
    H = "H"  # TODO declear


class BaseAccount:
    account_type: AccountType
    person_id: str
    broker_id: str
    account_id: str
    signed: bool = False

    def astype(self):
        return _ACCTTYPE.get(self.account_type, self.__class__)(**self.dict())


class Account(BaseAccount):
    username: str = ""
    
    def __init__(self, person_id, broker_id, account_id, signed=False, trader=''):
        self.person_id = person_id
        self.broker_id = broker_id
        self.account_id = account_id
        self.signed = signed
        self.trader = trader
    
    def __str__(self) -> str:
        return str(self.__dict__)

class StockAccount(Account):
    account_type: AccountType = AccountType.Stock


class FutureAccount(Account):
    account_type: AccountType = AccountType.Future


_ACCTTYPE = {"S": StockAccount, "F": FutureAccount}



from IntelligencePy import DT, MARKET_FLAG, RECOVER_STATUS, COM_STATUS
from PackagePy import *
import time

class Future:
    def __init__(self, callback = lambda dt, dic: print(dt, dic)):
        self.quoteCom = None
        self.callback = callback
        self.ProductList = {}
    
    def Future_RcvMessage(self, pkgDT:DT, pkg):
        if pkgDT == DT.QUOTE_I020:
            self.callback(pkgDT, PI20020(pkg))                
        elif pkgDT == DT.QUOTE_I021:                 
            self.callback(pkgDT, PI20021(pkg))
        elif pkgDT == DT.QUOTE_I030: 
            self.callback(pkgDT, PI20030(pkg))
        elif pkgDT == DT.QUOTE_I080: 
            self.callback(pkgDT, PI20080(pkg))
        elif pkgDT == DT.QUOTE_BASE_P08:
            # _PI20008 = PI20008(pkg)
            # self.callback(pkgDT, PI20008(pkg))
            # market = self.ProductList.get(_PI20008.Market, None)
            # if market is None:
            #     market = {}
            # market[_PI20008.Symbol] = _PI20008
            # self.ProductList[_PI20008.Market] = market
            # .append({ _PI20008.Symbol : _PI20008._PROD_NAME})
            # print(_PI20008)
            pass
        else:
            return False
        return True
    
    def Future_GetStatus(self, status:COM_STATUS, msg):
        if status == COM_STATUS.RECOVER_DATA:            
            # print('Future_GetStatus:[' + str(status) + ']:', msg)
            topic = str(msg)[1:]
            if topic not in self.future_RecoverStatus.keys():
                self.future_RecoverStatus[topic] = RECOVER_STATUS.RS_BEGIN
                print('開始回補', topic)
            else:
                self.future_RecoverStatus[topic] = RECOVER_STATUS.RS_DONE
                print('結束回補', topic)

    def Future_RecoverStatus(self, topic, status, count):
        status = RECOVER_STATUS(status)
        self.RecoverStatus[status] = { 'topic' : topic, 'count' : count }
        print(f"Future_RecoverStatus: [{status}] {self.RecoverStatus[status]}")

    def LoadTaifexProductXML(self):
        self.quoteCom.LoadTaifexProductXML()
        for i in range(3000):
            recover_status = self.RecoverStatus.get(RECOVER_STATUS.RS_DONE, None)
            if recover_status is not None:
                return RECOVER_STATUS.RS_DONE
            time.sleep(0.001)
        raise ValueError("download timeout")

    def GetTaifexProductBase(self, symbolId) -> PT01805:
        res = self.quoteCom.GetTaifexProductBase(symbolId)
        if res is None:
            raise ValueError("Need Download")
        return PT01805(res)

    def GetTaifexProductDetail(self, symbolId) -> PT01802:
        res = self.quoteCom.GetTaifexProductDetail(symbolId)
        if res is None:
            raise ValueError("Need Download")
        return PT01802(res)
    
    def RetriveQuoteList(self):
        self.future_RecoverStatus = {}
        res = self.quoteCom.RetriveQuoteList()
        if res < 0:
            errMsg = self.quoteCom.GetSubQuoteMsg(res)
            print(errMsg)
        time.sleep(0.1)
        double_Check = 0    #檢查是否有第二筆下載
        for _ in range(3000):
            if RECOVER_STATUS.RS_BEGIN in self.future_RecoverStatus.values():                
                double_Check = 0
            elif double_Check > 1:
                return 
            else:
                double_Check += 1
            time.sleep(0.001)
        raise ValueError("download timeout")
                

    def GetProdcutBase(self, symbolId) -> PI20008:
        res = self.quoteCom.GetProdcutBase(symbolId)
        if res is None:
            raise ValueError("Need Download")
        return PI20008(res)


from IntelligencePy import DT, RECOVER_STATUS
from PackagePy import *
import time
import random
from queue import Queue, Empty
from Intelligence import IdxKind

class Stock:
    def __init__(self, callback = lambda dt, dic: print(dt, dic)):
        self.quoteCom = None
        self.callback = callback
        self.RecoverStatus = {}
    
    def Stock_RcvMessage(self, pkgDT:DT, pkg):
        if pkgDT==DT.QUOTE_STOCK_MATCH1 or pkgDT==DT.QUOTE_STOCK_MATCH2 or \
            pkgDT==DT.QUOTE_ODD_MATCH1 or pkgDT==DT.QUOTE_ODD_MATCH2:
            self.callback(pkgDT, PI31001(pkg))
            pass
        elif pkgDT==DT.QUOTE_STOCK_DEPTH1 or pkgDT==DT.QUOTE_STOCK_DEPTH2 or \
                pkgDT==DT.QUOTE_ODD_DEPTH1 or pkgDT==DT.QUOTE_ODD_DEPTH2:
            self.callback(pkgDT, PI31002(pkg))
            pass
        elif pkgDT==DT.QUOTE_LAST_PRICE_STOCK or pkgDT==DT.QUOTE_LAST_PRICE_ODD:
            self.callback(pkgDT, PI30026(pkg))
            pass
        elif pkgDT==DT.QUOTE_LAST_INDEX1 or pkgDT==DT.QUOTE_LAST_INDEX2:
            _PI31026 = PI31026(pkg)
            if self.__LastIndexQueue:
                self.__LastIndexQueue.put(_PI31026)
            self.callback(pkgDT, _PI31026)
            pass
        elif pkgDT==DT.QUOTE_STOCK_INDEX1 or pkgDT==DT.QUOTE_STOCK_INDEX2:
            self.callback(pkgDT, PI31011(pkg))
            pass
        else:
            return False
        return True
    
    def Stock_RecoverStatus(self, topic, status, count):
        status = RECOVER_STATUS(status)
        self.RecoverStatus[status] = { 'topic' : topic, 'count' : count }
        # print(f"Stock_RecoverStatus: [{status}] {self.RecoverStatus[status]}")

    def RetriveProduct(self, market="TSE"):
        match market:
            case "TSE":
                status = self.quoteCom.RetriveProductTSE()
            case "OTC":
                status = self.quoteCom.RetriveProductOTC() 

        if (status<0): 
            errmsg= self.quoteCom.GetSubQuoteMsg(status)
            raise ValueError(f"'RetriveProduct' market: {market}, Error: {errmsg}")

        for i in range(3000):
            recover_status = self.RecoverStatus.get(RECOVER_STATUS.RS_DONE, None)
            if recover_status is not None:
                return RECOVER_STATUS.RS_DONE
            time.sleep(0.001)
        raise ValueError(f"'RetriveProduct' market: {market}, timeout")
    
    def RetriveLastIndex(self, market="TSE"):
        self.__LastIndexQueue = Queue()     
        match market:
            case "TSE":
                status = self.quoteCom.RetriveLastIndex(IdxKind.IdxKind_List)
            case "OTC":
                status = self.quoteCom.RetriveLastIndex(IdxKind.IdxKind_OTC)         
            case _:
                raise TypeError("market type is wrong.")
            
        if (status<0): 
            errmsg= self.quoteCom.GetSubQuoteMsg(status)            
            raise ValueError(f"'RetriveLastIndex' market: {market}, Error: {errmsg}")

        try:
            pkg:PI31026 = self.__LastIndexQueue.get()
            return pkg
        except Empty:
            raise ValueError(f"'RetriveLastIndex' market: {market}, timeout")

        

    def GetProductList(self, symbolId=None, market="TSE"):
        symbols:list[str]
        match market:
            case "TSE":
                symbols = self.quoteCom.GetProductListTSE()#代碼|名稱|漲停價|參考價|跌停價|上次成交日
            case "OTC":
                symbols = self.quoteCom.GetProductListOTC()#代碼|名稱|漲停價|參考價|跌停價|上次成交日
        
        
        start = time.process_time()
        # for stock in symbols:
        #     temp = stock.split('|')
        #     symbolId = temp[0]
        #     self.GetProductSTOCK(symbolId)
        

        symbolIds = []
        while(len(symbolIds) < 10):
            temp = random.choice(symbols).split('|')
            symbolId = temp[0]
            # if len(symbolId) != 4:
            #     continue
            symbolIds.append(symbolId)
            # Product = self.GetProductSTOCK(symbolId)
            # StockInfo = self.StockInfo(temp)
            # if Product.Bull_Price != StockInfo['Bull_Price'] or \
            #     Product.Ref_Price != StockInfo['Ref_Price'] or \
            #     Product.Bear_Price != StockInfo['Bear_Price'] :
            #     print(Product.StockNo, StockInfo['StockNo'])

        end = time.process_time()
        print("GetProductListTSC process_time 測量時間：%f 秒" % (end - start))
        return symbolIds

    def GetProductSTOCK(self,symbolId) -> PI30001:
        res = self.quoteCom.GetProductSTOCK(symbolId)
        if res is None:            
            raise ValueError("Need Download")
        return PI30001(res)

    def StockInfo(self, pkg):
        #代碼|名稱|漲停價|參考價|跌停價|上次成交日
        return {
            'StockNo' : pkg[0],
            'StockName' : pkg[1],            
            'Bull_Price' : float(str(pkg[2])),
            'Ref_Price' : float(str(pkg[3])),
            'Bear_Price' : float(str(pkg[4])),
            'LastTradeDate' : pkg[5]
        }


import sys, os
import clr
from queue import Queue, Empty
import datetime

assembly_path = os.path.dirname(__file__)
sys.path.append(assembly_path)
clr.AddReference("QuoteCom")     #必要引用dll

from Intelligence import QuoteCom   #from namespace import class
from IntelligencePy import DT 
from IntelligencePy import COM_STATUS, RECOVER_STATUS
from PackagePy import *
from QuoteCom.Login import Login
from QuoteCom.Stock import Stock
from QuoteCom.Future import Future

class Quote(Login, Stock, Future):
    """KGI期貨國內報價的Python API範例程式。
    """
    def __init__(self, host, port, sid, token, callback = lambda dt, dic: print(dt, dic)) -> None:
        """程式初始化

        Args:
            host (str): 主機連線的host
            port (num): 主機連線的port
            sid (str):  主機連線的sid
            token (str): 主機連線的token
        """
        Login.__init__(self, host, port)
        Future.__init__(self)
        Stock.__init__(self)
        self.sid = sid
        self.token = token
        self.callback = callback        
        self.quoteCom = QuoteCom("", port, sid, token)
        print("QuoteCom API 初始化 Version (%s) ........" % (self.quoteCom.version))
        # register event handler
        #資料回補事件KGI QuoteCom API status event
        self.quoteCom.OnRecoverStatus += self.onQuoteRecoverStatus
        #狀態通知事件KGI QuoteCom API message event
        self.quoteCom.OnRcvMessage += self.onQuoteRcvMessage
        #資料接收事件KGI QuoteCom API status event
        self.quoteCom.OnGetStatus += self.onQuoteGetStatus

        self.com_status = None
        self.Com_StatusDict = {}
        self.GetStatusQueue = Queue()
        self.RcvMessageQueue = Queue()
        # self.DTlist = []

    #region 回報
    def onQuoteRecoverStatus(self, sender, topic, status, count):
        try:
            print("onQuoteRecoverStatus: [{topic}] {status} , {count}".format(topic=topic,status=status,count=count))
            status = RECOVER_STATUS(status)
            self.RecoverStatus[status] = { 'topic' : topic, 'count' : count }
            # self.Stock_RecoverStatus(topic, status, count)
            # self.Future_RecoverStatus(topic, status, count)
        except Exception as e:
            print(e)  

    def onQuoteRcvMessage(self, sender, pkg):
        try:
            pkgDT = DT(pkg.DT)
            if self.Login_RcvMessage(pkgDT, pkg):
                pass
            elif self.Stock_RcvMessage(pkgDT, pkg):
                pass
            # elif self.Future_RcvMessage(pkgDT, pkg):
            #     pass
            else:
                self.callback(pkgDT, PackageBase(pkg))
                # if pkgDT not in self.DTlist:
                #     self.DTlist.append(pkgDT)
                #     print(self.DTlist)
                #     PackageCode(pkg)
                
                # print("onTradeRcvMessage: DT=[{DT}({dt})] {Log} ".format(dt=pkg.DT,DT=DT(pkg.DT),Log=pkg.ToLog()))

                # now = datetime.datetime.now()
                # formatted_time = "{:02}{:02}{:02}{:03}".format(now.hour, now.minute, now.second, now.microsecond//1000)
                # time_int = int(formatted_time)
                # diff = time_int - pkg.Match_Time
                # self.callback("1.[{stock}] 資料時間:[{time} - {now}] = {diff}".format(stock=pkg.StockNo,time=pkg.Match_Time,now=time_int,diff=diff))

                pass
        except Exception as e:
            print(e)  
    
    def onQuoteGetStatus(self, sender, status, msg) :
        try:
            self.com_status = COM_STATUS(status)
            smsg = bytes(msg).decode('UTF-8','strict')
            self.Com_StatusDict[self.com_status] = { 'Time' : datetime.datetime.now() , 'Msg' : smsg }
            # print('onTradeGetStatus:[' + str(status) + ']:' + smsg)
            self.Login_GetStatus(self.com_status, smsg)
            self.Future_GetStatus(self.com_status, smsg)
        except Exception as e:
            print(e)
    #endregion



import sys, os, clr
from enum import Enum

assembly_path = os.path.dirname(__file__)
sys.path.append(assembly_path)
clr.AddReference("TradeCom")     #必要引用dll

from Intelligence import DT as _DT

#自動產出Enum
"""for key in _DT.__dict__.keys():
    value = _DT.__dict__.get(key)
    if 'Intelligence.DT' in str(value) and "__" not in key:
        print("{key} = int(_DT.{key})".format(key=key))"""

class DT(Enum):
    MESSAGE_MAP = int(_DT.MESSAGE_MAP)
    BROKER_MAP = int(_DT.BROKER_MAP)
    BROKER_NAME = int(_DT.BROKER_NAME)
    AP_VERSION = int(_DT.AP_VERSION)
    LOGIN = int(_DT.LOGIN)
    AUTH_EXCHANGE = int(_DT.AUTH_EXCHANGE)
    PreviousLogin = int(_DT.PreviousLogin)
    LOGIN_AGENT = int(_DT.LOGIN_AGENT)
    CUSTOMER_AGENT = int(_DT.CUSTOMER_AGENT)
    HK_CHANGE_PWD = int(_DT.HK_CHANGE_PWD)
    INVENTORY_YESTERDAY = int(_DT.INVENTORY_YESTERDAY)
    FINANCIAL = int(_DT.FINANCIAL)
    INVENTORY = int(_DT.INVENTORY)
    INVENTORY_DETAIL = int(_DT.INVENTORY_DETAIL)
    FINANCIAL_COVER = int(_DT.FINANCIAL_COVER)
    FINANCIAL_COVER_TRADER = int(_DT.FINANCIAL_COVER_TRADER)
    INVENTORY_DETAIL_TRADER = int(_DT.INVENTORY_DETAIL_TRADER)
    INVENTORY_TRADER = int(_DT.INVENTORY_TRADER)
    FINANCIAL_TRADERN = int(_DT.FINANCIAL_TRADERN)
    FINANCIAL_TRADERNN = int(_DT.FINANCIAL_TRADERNN)
    FINANCIAL_STRIKE = int(_DT.FINANCIAL_STRIKE)
    FINANCIAL_COVERDH = int(_DT.FINANCIAL_COVERDH)
    FINANCIAL_RECIPROCATE = int(_DT.FINANCIAL_RECIPROCATE)
    FINANCIAL_CURRENCY = int(_DT.FINANCIAL_CURRENCY)
    FINANCIAL_SERVICECHARGE = int(_DT.FINANCIAL_SERVICECHARGE)
    FINANCIAL_TPASS = int(_DT.FINANCIAL_TPASS)
    FINANCIAL_APPLYWITHDRAW = int(_DT.FINANCIAL_APPLYWITHDRAW)
    FINANCIAL_APPLYWITHDRAW_PWD = int(_DT.FINANCIAL_APPLYWITHDRAW_PWD)
    FINANCIAL_WITHDRAW = int(_DT.FINANCIAL_WITHDRAW)
    NOTICE_RPT = int(_DT.NOTICE_RPT)
    NOTICE = int(_DT.NOTICE)
    INVENTORY_YESTERDAY_Detail = int(_DT.INVENTORY_YESTERDAY_Detail)
    FINANCIAL_COVER_TRADER_Detail = int(_DT.FINANCIAL_COVER_TRADER_Detail)
    FUT_PRODUCT_INFO = int(_DT.FUT_PRODUCT_INFO)
    FUT_ORDER_ACK = int(_DT.FUT_ORDER_ACK)
    FINCIAL_ORDER_RCV = int(_DT.FINCIAL_ORDER_RCV)
    FUT_ORDER_ACK_AGENT = int(_DT.FUT_ORDER_ACK_AGENT)
    FUT_ORDDER_CHANGE_ACK = int(_DT.FUT_ORDDER_CHANGE_ACK)
    FUT_ORDER_RPT = int(_DT.FUT_ORDER_RPT)
    FUT_ORDER_RPT_K = int(_DT.FUT_ORDER_RPT_K)
    FUT_DEAL_RPT = int(_DT.FUT_DEAL_RPT)
    FUT_DEAL_RPT_K = int(_DT.FUT_DEAL_RPT_K)
    FUT_RPT_RECOVER = int(_DT.FUT_RPT_RECOVER)
    FUT_RPT_RECOVER_TRADER = int(_DT.FUT_RPT_RECOVER_TRADER)
    FUT_RPT_RECOVER_AGENT = int(_DT.FUT_RPT_RECOVER_AGENT)
    FUT_RPT_RECOVER_KTR = int(_DT.FUT_RPT_RECOVER_KTR)
    FFUT_ORDER_ACK = int(_DT.FFUT_ORDER_ACK)
    FFUT_ORDER_ACK2 = int(_DT.FFUT_ORDER_ACK2)
    FFUT_ORDER_RPT = int(_DT.FFUT_ORDER_RPT)
    FFUT_DEAL_RPT = int(_DT.FFUT_DEAL_RPT)
    FFUT_ORDER_RECOVER = int(_DT.FFUT_ORDER_RECOVER)
    FFUT_GTC_RPT = int(_DT.FFUT_GTC_RPT)
    CONDITION_SET = int(_DT.CONDITION_SET)
    CONDITION_MOD = int(_DT.CONDITION_MOD)
    CONDITION_QRY = int(_DT.CONDITION_QRY)
    CONDITION_NOTIFY = int(_DT.CONDITION_NOTIFY)
    SECU_ORDER_ACK = int(_DT.SECU_ORDER_ACK)
    SECU_ORDER_RPT = int(_DT.SECU_ORDER_RPT)
    SECU_DEAL_RPT = int(_DT.SECU_DEAL_RPT)
    SECU_RPT_RECOVER = int(_DT.SECU_RPT_RECOVER)
    SECU_ORDER_ACK_N = int(_DT.SECU_ORDER_ACK_N)
    SECU_SUBORDER_ACK = int(_DT.SECU_SUBORDER_ACK)
    SECU_SUBORDER_RPT = int(_DT.SECU_SUBORDER_RPT)
    SECU_SUBDEAL_RPT = int(_DT.SECU_SUBDEAL_RPT)
    FINANCIAL_RCORDER = int(_DT.FINANCIAL_RCORDER)
    FINANCIAL_RCMATCHSUM = int(_DT.FINANCIAL_RCMATCHSUM)
    FINANCIAL_RCMATCHDETAIL = int(_DT.FINANCIAL_RCMATCHDETAIL)
    FINANCIAL_RCPOSITIONSUM = int(_DT.FINANCIAL_RCPOSITIONSUM)
    FINANCIAL_RCCURRENCY = int(_DT.FINANCIAL_RCCURRENCY)
    FINANCIAL_RCSTOCKPOSITION = int(_DT.FINANCIAL_RCSTOCKPOSITION)
    FINANCIAL_RCDELIVERY = int(_DT.FINANCIAL_RCDELIVERY)
    FINANCIAL_WSSETAMTTRIAL = int(_DT.FINANCIAL_WSSETAMTTRIAL)
    FINANCIAL_WSSETAMTDETAIL = int(_DT.FINANCIAL_WSSETAMTDETAIL)
    FINANCIAL_WSSETTLEAMT = int(_DT.FINANCIAL_WSSETTLEAMT)
    FINANCIAL_WSINVENTORY = int(_DT.FINANCIAL_WSINVENTORY)
    FINANCIAL_WSINVENTORYSUM = int(_DT.FINANCIAL_WSINVENTORYSUM)
    FINANCIAL_WSBALANCESTATEMENT = int(_DT.FINANCIAL_WSBALANCESTATEMENT)
    FINANCIAL_WSREALIZEPL = int(_DT.FINANCIAL_WSREALIZEPL)
    SECU_CREDITINFO = int(_DT.SECU_CREDITINFO)
    SECU_EARMARK_SET = int(_DT.SECU_EARMARK_SET)
    SECU_ALLOWANCE_RPT = int(_DT.SECU_ALLOWANCE_RPT)
    SECU_ALLOWANCE = int(_DT.SECU_ALLOWANCE)
    SECU_TRANSFER_APPLY = int(_DT.SECU_TRANSFER_APPLY)
    SECU_TRANSFER_QRY = int(_DT.SECU_TRANSFER_QRY)
    SECU_TRANSFER_INV = int(_DT.SECU_TRANSFER_INV)
    # QUOTE_BASE_P08 = int(_DT.QUOTE_BASE_P08)
    # QUOTE_BASE_P08_PM = int(_DT.QUOTE_BASE_P08_PM)
    QUOTE_CLOSE_I070 = int(_DT.QUOTE_CLOSE_I070)
    QUOTE_I020 = int(_DT.QUOTE_I020)
    QUOTE_I024 = int(_DT.QUOTE_I024)
    QUOTE_I022 = int(_DT.QUOTE_I022)
    QUOTE_I021 = int(_DT.QUOTE_I021)
    QUOTE_I023 = int(_DT.QUOTE_I023)
    QUOTE_I030 = int(_DT.QUOTE_I030)
    QUOTE_I060 = int(_DT.QUOTE_I060)
    QUOTE_I080 = int(_DT.QUOTE_I080)
    QUOTE_I081 = int(_DT.QUOTE_I081)
    QUOTE_I090 = int(_DT.QUOTE_I090)
    QUOTE_I082 = int(_DT.QUOTE_I082)
    QUOTE_LAST_PRICE = int(_DT.QUOTE_LAST_PRICE)
    QUOTE_INFO = int(_DT.QUOTE_INFO)
    QUOTE_SESSION = int(_DT.QUOTE_SESSION)
    QUOTE_I101 = int(_DT.QUOTE_I101)
    QUOTE_STOCK_T30 = int(_DT.QUOTE_STOCK_T30)
    QUOTE_STOCK_MATCH1 = int(_DT.QUOTE_STOCK_MATCH1)
    QUOTE_STOCK_MATCH2 = int(_DT.QUOTE_STOCK_MATCH2)
    QUOTE_STOCK_DEPTH1 = int(_DT.QUOTE_STOCK_DEPTH1)
    QUOTE_STOCK_DEPTH2 = int(_DT.QUOTE_STOCK_DEPTH2)
    QUOTE_ODD_MATCH1 = int(_DT.QUOTE_ODD_MATCH1)
    QUOTE_ODD_MATCH2 = int(_DT.QUOTE_ODD_MATCH2)
    QUOTE_ODD_DEPTH1 = int(_DT.QUOTE_ODD_DEPTH1)
    QUOTE_ODD_DEPTH2 = int(_DT.QUOTE_ODD_DEPTH2)
    QUOTE_LAST_PRICE_ODD = int(_DT.QUOTE_LAST_PRICE_ODD)
    QUOTE_STOCK_TREND1 = int(_DT.QUOTE_STOCK_TREND1)
    QUOTE_STOCK_TREND2 = int(_DT.QUOTE_STOCK_TREND2)
    QUOTE_ODD_TREND1 = int(_DT.QUOTE_ODD_TREND1)
    QUOTE_ODD_TREND2 = int(_DT.QUOTE_ODD_TREND2)
    QUOTE_STOCK_INDEX1 = int(_DT.QUOTE_STOCK_INDEX1)
    QUOTE_STOCK_INDEX2 = int(_DT.QUOTE_STOCK_INDEX2)
    QUOTE_STOCK_NEWINDEX1 = int(_DT.QUOTE_STOCK_NEWINDEX1)
    QUOTE_STOCK_NEWINDEX2 = int(_DT.QUOTE_STOCK_NEWINDEX2)
    QUOTE_STOCK_ORDSUM1 = int(_DT.QUOTE_STOCK_ORDSUM1)
    QUOTE_STOCK_ORDSUM2 = int(_DT.QUOTE_STOCK_ORDSUM2)
    QUOTE_STOCK_AVGINDEX = int(_DT.QUOTE_STOCK_AVGINDEX)
    QUOTE_LAST_INDEX1 = int(_DT.QUOTE_LAST_INDEX1)
    QUOTE_LAST_INDEX2 = int(_DT.QUOTE_LAST_INDEX2)
    QUOTE_LAST_PRICE_STOCK = int(_DT.QUOTE_LAST_PRICE_STOCK)
    QUOTE_STOCK_RECOVERY = int(_DT.QUOTE_STOCK_RECOVERY)
    QUOTE_STOCK_MATCH3 = int(_DT.QUOTE_STOCK_MATCH3)
    FFUT_ORDER_RPT2 = int(_DT.FFUT_ORDER_RPT2)
    FFUT_DEAL_RPT2 = int(_DT.FFUT_DEAL_RPT2)
    FFUT_KRPT_RECOVER3 = int(_DT.FFUT_KRPT_RECOVER3)
    FFUT_RPT_RECOVER3 = int(_DT.FFUT_RPT_RECOVER3)
    FFUT_ORDER_RPT3 = int(_DT.FFUT_ORDER_RPT3)
    FFUT_DEAL_RPT3 = int(_DT.FFUT_DEAL_RPT3)
    KTRADER_CDAUTHORITY = int(_DT.KTRADER_CDAUTHORITY)
    FFUT_RPT_RECOVER = int(_DT.FFUT_RPT_RECOVER)
    FFUT_RPT_RECOVER_TRADER = int(_DT.FFUT_RPT_RECOVER_TRADER)
    FFUT_RPT_RECOVER_SPREAD = int(_DT.FFUT_RPT_RECOVER_SPREAD)
    FFUT_RPT_RECOVER_TRADER_SPREAD = int(_DT.FFUT_RPT_RECOVER_TRADER_SPREAD)
    AES_TEST = int(_DT.AES_TEST)
    QUOTE_I080_RECOVER = int(_DT.QUOTE_I080_RECOVER)
    QUOTE_I020_RECOVER = int(_DT.QUOTE_I020_RECOVER)
    QUOTE_RECOVER = int(_DT.QUOTE_RECOVER)
    QUOTE_FOREIGN_PRECOVER = int(_DT.QUOTE_FOREIGN_PRECOVER)
    QUOTE_FOREIGN_PRODUCT = int(_DT.QUOTE_FOREIGN_PRODUCT)
    QUOTE_FOREIGN_SESSION_STATUS = int(_DT.QUOTE_FOREIGN_SESSION_STATUS)
    QUOTE_FOREIGN_MATCH = int(_DT.QUOTE_FOREIGN_MATCH)
    QUOTE_FOREIGN_CLOSEINFO = int(_DT.QUOTE_FOREIGN_CLOSEINFO)
    QUOTE_FOREIGN_SETTLEMENTPRICE = int(_DT.QUOTE_FOREIGN_SETTLEMENTPRICE)
    QUOTE_FOREIGN_TOTALQTY = int(_DT.QUOTE_FOREIGN_TOTALQTY)
    QUOTE_FOREIGN_INDEX = int(_DT.QUOTE_FOREIGN_INDEX)
    QUOTE_FOREIGN_DEPTH = int(_DT.QUOTE_FOREIGN_DEPTH)
    QUOTE_FOREIGN_DEPTH_SS = int(_DT.QUOTE_FOREIGN_DEPTH_SS)
    QUOTE_FOREIGN_LASTPRICE = int(_DT.QUOTE_FOREIGN_LASTPRICE)
    SSO_TRUSTCODE = int(_DT.SSO_TRUSTCODE)


import typing
from TradeCom import TradeCom
from IntelligencePy import *
from PackagePy import *
from constant import OrderState

class OrderReport:
    def __init__(self, tradeCom) -> None:
        self.__tradeCom : TradeCom = tradeCom
        self.__tradeCom.callback = self.__tradeCom_callback
        self.order_callback = None
    
    def set_order_callback(
        self, func: typing.Callable[[OrderState, dict], None]
    ) -> None:
        self.order_callback = func

    def __tradeCom_callback(self, pkgDT:DT, pkg):
        if self.order_callback == None:
            return

        match pkgDT :
            case DT.SECU_ORDER_RPT:
                self.order_callback(OrderState.StockOrder, StockOrder(pkg).__dict__)
                return
            case DT.SECU_DEAL_RPT:
                self.order_callback(OrderState.StockDeal, StockDeal(pkg).__dict__)
                return
            case DT.FUT_ORDER_RPT:
                self.order_callback(OrderState.FuturesOrder, FuturesOrder(pkg).__dict__)
                return
            case DT.FUT_DEAL_RPT:                
                self.order_callback(OrderState.FuturesDeal, FuturesDeal(pkg).__dict__)
                return
            case _:
                # print(pkgDT, pkg)
                return

#region OrderRPT

class PackageConvert:
    def op_type(OrderFunc):
        return { "I": "New", "D": "Cancel", "M": "UpdatePrice", "R": "UpdatePrice", "C": "UpdateQty" }.get(OrderFunc, None)
    
    def action(Side):
        return { 'B' : "Buy" , 'S' : "Sell"}.get(Side, None)
    
    def order_type(TimeInForceN):
        return { 'F' : "FOK" , 'I' : "IOC" , 'R' : "ROD" }.get(TimeInForceN, None)
    
    def price_type(PriceFlagN): 
        return { '1' : "MKT" , '2' : "LMT" , '0' : "MKP" }.get(PriceFlagN, None)
    
    def order_cond(OrdClass): 
        return { '0' : "CASH", '3' : "MarginTrading", '4' : "ShortSelling", '6' : "ShortUnLimit", 'A' : "EDN" }.get(OrdClass, None)
    
    def order_lot(OrdLot):
        return { '0' : "Common", '2' : "Fixing", '1' : "Odd", '4' : "IntradayOdd" }.get(OrdLot, None)
    
    def oc_type(PositionEffect):
        return { 'O':"New", 'C':"Cover", 'A':"Auot" }.get(PositionEffect, None)

class Operation:
    op_type: (str) #"New": 新單, "Cancel": 刪單, "UpdatePrice": 改價, "UpdateQty": 改量
    op_code: (str) #{"00": 成功, others: 失敗}
    op_msg: (str) #錯誤訊息

class Order:
    id: (str)# 與成交回報的trade_id相同
    seqno: (str)# 平台單號
    ordno: (str)# 委託單號
    account: (dict)# 帳號資訊
    action: (str)# 買賣別 {Buy, Sell}
    price: (float)# 委託價格
    quantity: (int)# 委託數量
    order_type: (str)# 委託類別 {ROD, IOC, FOK}
    price_type: (str)# {LMT: 限價, MKT: 市價, MKP: 範圍市價}
    order_cond: (str)# {Cash: 現股, MarginTrading: 融資, ShortSelling: 融券, ShortUnLimit:借券賣出, EDN:賣出擔保品 }
    #證券
    order_lot: (str)# { Common: 整股, Fixing: 定盤, Odd: 盤後零股, IntradayOdd: 盤中零股 }
    custom_field: (str)# 自訂欄位
    #期貨
    market_type : (str) # 市場別 {Day:日盤, Night:夜盤}
    oc_type : (str) # {New: 新倉, Cover: 平倉, Auto: 自動}
    subaccount : (str) # 子帳號
    combo : (bool) # 是否為組合單

class Status:
    id: (str)#: 與成交回報的trade_id相同
    exchange_ts: (int)#: 交易所時間
    modified_price: (float)#: 改價
    cancel_quantity:(int)#: 取消數量
    order_quantity: (int)#: 委託數量
    web_id: (str)#: 下單平台代碼

class Contract:
    security_type: (str)# 商品類別
    exchange: (str)# 交易所
    code: (str)# 商品代碼
    #證券
    symbol: (str)# 符號
    name: (str)# 商品名稱
    currency: (str)# 幣別
    #期貨
    delivery_month : (str) # 交割月份
    delivery_date : (str) # 交割日期
    strike_price : (float) # 履約價
    option_right : (str) # {Future, OptionCall, OptionPut}

class OrderRPT:
    operation: (dict)
    order: (dict)
    status: (dict)
    contract: (dict)

    def __str__(self) -> str:
        return str(self.__dict__)

class StockOrder(OrderRPT):
    def __init__(self, pkg:PT04010):
        # print(pkg)
        operation = Operation()
        operation.op_type = PackageConvert.op_type(pkg.OrderFunc)
        operation.op_code = pkg.ErrCode
        operation.op_msg = pkg.ErrMsg    
        self.operation = operation.__dict__
        
        order = Order()
        order.id = pkg.CNTN
        # order.seqno 
        order.ordno = pkg.OrderNo
        order.account= {
            'account_type': 'S', 
            'broker_id': pkg.BrokerId, 
            'account_id': pkg.Account, 
        }
        order.action = PackageConvert.action(pkg.Side)
        order.price = pkg.Price
        order.quantity = pkg.Qty
        order.order_type = PackageConvert.order_type(pkg.TimeInForceN)
        order.price_type = PackageConvert.price_type(pkg.PriceFlagN)
        order.order_cond = PackageConvert.order_cond(pkg.OrdClass)
        order.order_lot = PackageConvert.order_lot(pkg.OrdLot)
        # order.custom_field = 
        self.order = order.__dict__

        status = Status()
        status.id = pkg.CNTN
        # status.exchange_ts
        status.modified_price = pkg.Price if pkg.OrderFunc == 'M' else 0
        status.cancel_quantity = pkg.Qty if pkg.OrderFunc in ['C','D'] else 0
        status.order_quantity = pkg.Qty if pkg.OrderFunc == 'I' else 0
        # status.web_id
        self.status = status.__dict__

        contract = Contract()
        contract.security_type = "STK"
        contract.exchange = { '0' : "TSE" , '1' : "OTC" }.get(pkg.Market, None)
        contract.code = pkg.StockID
        contract.symbol = ''
        contract.name = ''
        contract.currency = 'TWD'
        self.contract = contract.__dict__

class FuturesOrder(OrderRPT):
    def __init__(self, pkg:PT02010):
        operation = Operation()
        operation.op_type = PackageConvert.op_type(pkg.OrderFunc)
        operation.op_code = pkg.Code
        operation.op_msg = pkg.ErrMsg    
        self.operation = operation.__dict__

        order = Order()
        order.id = pkg.CNT
        # order.seqno 
        order.ordno = pkg.OrderNo
        order.account= {
            'account_type': 'F', 
            'broker_id': pkg.BrokerId, 
            'account_id': pkg.Account, 
        }
        order.action = PackageConvert.action(pkg.Side)
        order.price = pkg.Price
        order.quantity = pkg.AfterQty
        order.order_type = PackageConvert.order_type(pkg.TimeInForce)
        order.price_type = PackageConvert.price_type(pkg.PriceMark)
        # order.market_type 
        order.oc_type = PackageConvert.oc_type(pkg.PositionEffect)
        # order.subaccount
        # order.combo
        self.order = order.__dict__

        status = Status()
        status.id = pkg.CNT
        # status.exchange_ts
        status.modified_price = pkg.Price if pkg.OrderFunc == 'R' else 0
        status.cancel_quantity = { 'D' : pkg.BeforeQty, 'C' : int(pkg.BeforeQty)-int(pkg.AfterQty) }.get(pkg.OrderFunc, 0)
        status.order_quantity = pkg.AfterQty if pkg.OrderFunc == 'I' else 0
        status.web_id = pkg.WebID
        self.status = status.__dict__

        contract = Contract()
        contract.security_type = "FUT"
        contract.exchange = "TAIFEX"
        contract.code = pkg.Symbol
        # contract.delivery_month
        # contract.delivery_date
        # contract.strike_price
        # contract.option_right
        self.contract = contract.__dict__

#endregion

class DealRPT:
    trade_id : (str) # 與委託回報id相同
    seqno : (str) # 平台單號
    ordno : (str) # 前五碼為同委託回報委託單號，後三碼為同筆委託成交交易序號。
    exchange_seq : (str) # 回報序號
    broker_id : (str) # 分行代碼
    account_id : (str) # 帳號
    action : (str) # 買賣別 {Buy, Sell}
    code : (str) # 商品代碼
    price : (float) # 成交價
    quantity : (int) # 成交量
    ts : (int) # 成交時間戳
    
    def __str__(self) -> str:
        return str(self.__dict__)

class StockDeal(DealRPT):
    order_cond : (str) # { Cash: 現股, MarginTrading: 融資, ShortSelling: 融券 }
    order_lot : (str) # { Common: 整股, Fixing: 定盤, Odd: 盤後零股, IntradayOdd: 盤中零股 }
    web_id : (str) # 平台代碼
    custom_field : (str) # 自訂欄位

    def __init__(self, pkg:PT04011):
        self.trade_id = pkg.CNTN
        # self.seqno
        self.ordno = pkg.OrderNo
        # self.exchange_seq
        self.broker_id = pkg.BrokerId
        self.account_id = pkg.Account
        self.action = PackageConvert.action(pkg.Side)
        self.code = pkg.StockID
        self.order_cond = PackageConvert.order_cond(pkg.OrdClass)
        self.order_lot = PackageConvert.order_lot(pkg.OrdLot)
        self.price = pkg.PriceN
        self.quantity = pkg.DealQtyN
        # self.web_id
        # self.custom_field
        # self.ts

class FuturesDeal(DealRPT):
    subaccount : (str) # 子帳號
    security_type : (str) # 商品類別
    delivery_month : (str) # 交割月份
    strike_price : (float) # 履約價
    option_right : (str) # {Future, OptionCall, OptionPut}
    market_type : (str) # {Day, Night}

    def __init__(self, pkg:PT02011):
        # print(pkg)
        self.trade_id = pkg.CNT
        # self.seqno
        self.ordno = pkg.OrderNo
        self.exchange_seq = pkg.MarketNo
        self.broker_id = pkg.BrokerId
        self.account_id = pkg.Account
        self.action = PackageConvert.action(pkg.Side)
        self.code = pkg.Symbol
        self.price = pkg.DealPrice
        self.quantity = pkg.DealQty
        # self.subaccount
        self.security_type = pkg.Market
        # self.delivery_month 
        # self.strike_price
        # self.option_right
        # self.market_type
        # self.ts


import typing
import datetime

from account import Account
from contracts import Contract
from constant import (
    Action,
    FuturesPriceType,
    OrderType,
    FuturesOCType,
    StockPriceType,
    StockOrderCond,
    StockOrderLot,
    Status,
    OrderState,
)

class BaseModel:
    def __str__(self) -> str:
        return str(self.__dict__)

class Deal(BaseModel):
    seq: str
    price: typing.Union[float]
    quantity: int
    ts: float


class OrderStatus(BaseModel):
    id: str = ""
    status: Status
    status_code: str = ""
    web_id: str = ""
    order_datetime: typing.Optional[datetime.datetime] = None
    msg: str = ""
    modified_time: typing.Optional[datetime.datetime] = None
    modified_price: typing.Union[float] = 0
    order_quantity: int = 0
    deal_quantity: int = 0
    cancel_quantity: int = 0
    deals: typing.List[Deal] = None

class BaseOrder(BaseModel):
    action: Action
    price: typing.Union[float]
    quantity: int
    id: str = ""
    seqno: str = ""
    ordno: str = ""
    account: Account = None
    custom_field: str = ""
    ca: str = ""

    # def __repr_args__(self):
    #     return [
    #         (k, v)
    #         for k, v in self._iter(to_dict=False, exclude_defaults=True, exclude={"ca"})
    #     ]

    def __init__(self, action: Action, price: typing.Union[float], quantity: int, **kwargs):
        self.action = action
        self.price = price
        self.quantity = quantity
        # self.id = ""
        # self.seqno = ""
        # self.ordno = ""
        self.account = kwargs.get("account",None)
        self.custom_field = kwargs.get("custom_field","")
        # self.ca = ""

class FuturesOrder(BaseOrder):
    price_type: FuturesPriceType
    order_type: OrderType
    octype: FuturesOCType = FuturesOCType.Auto

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        price_type = kwargs.get('price_type')
        if not isinstance(price_type, FuturesPriceType):
            return
        self.price_type = price_type
        self.order_type = kwargs.get('order_type')
        self.octype = kwargs.get('octype', self.octype)
                

class StockOrder(BaseOrder):
    price_type: StockPriceType
    order_type: OrderType
    order_lot: StockOrderLot = StockOrderLot.Common
    order_cond: StockOrderCond = StockOrderCond.Cash
    daytrade_short: bool = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        price_type = kwargs.get('price_type')
        if not isinstance(price_type, StockPriceType):
            return
        self.price_type = price_type
        self.order_type = kwargs.get('order_type')
        self.order_lot = kwargs.get('order_lot', self.order_lot)
        self.order_cond = kwargs.get('order_cond', self.order_cond)
        self.daytrade_short = kwargs.get('daytrade_short', False)
        

class Order(StockOrder, FuturesOrder):
    price_type: typing.Union[StockPriceType, FuturesPriceType]
    order_type: OrderType

    def __init__(
        self,
        price: typing.Union[float],
        quantity: int,
        action: Action,
        price_type: typing.Union[StockPriceType, FuturesPriceType],
        order_type: OrderType,
        **kwargs
    ):
        super().__init__(
            **{
                **dict(
                    price=price,
                    quantity=quantity,
                    action=action,
                    price_type=price_type,
                    order_type=order_type,
                ),
                **kwargs,
            }
        )

OrderTypeVar = typing.Union[Order, StockOrder, FuturesOrder]

class Trade:
    contract: Contract
    order: OrderTypeVar
    status: OrderStatus

    def __init__(self, contract: Contract, order: OrderTypeVar, status: OrderStatus):
        # super().__init__(**dict(contract=contract, order=order, status=status))
        self.contract=contract
        self.order=order
        self.status=status




from IntelligencePy import DT
from IntelligencePy import COM_STATUS
from PackagePy import *
from queue import Queue, Empty
import datetime

class Login:
    def __init__(self, host, port):
        self.tradecom = None
        self.host = host
        self.port = port
        self.accounts = []

        self.__com_status = None
        self.__Com_StatusDict = dict()
        self.__GetStatusQueue = Queue()
        self.__RcvMessageQueue = Queue()

    def Dispose(self):
        """關閉API的元件
        """
        self.tradecom.Dispose()

    def Logout(self):
        """登出API平台
        """
        self.tradecom.Logout()

    def Login(self, uid, pwd, subscribe_report = True, recover_report = True, loginInfos=None):
        """自行登入
        Args:
            uid (_type_): _description_
            pwd (_type_): _description_
        """
        #是否註冊即時回報
        self.tradecom.AutoSubReport = subscribe_report
        self.tradecom.AutoSubReportSecurity = subscribe_report
        #是否回補回報
        self.tradecom.AutoRecoverReport = recover_report
        self.tradecom.AutoRecoverReportSecurity = recover_report

        if loginInfos is None:
            loginInfos = f'{uid},,{pwd}'
        self.tradecom.LoginDirect(self.host, self.port, loginInfos)
        try:
            accountInfos = loginInfos.split("|") #check number of accounts
            for account in accountInfos:
                if len(account) == 0:   #filter empty account
                    continue
                status = self.__GetStatusQueue.get(timeout=5)
                # print(status, self.Com_StatusDict[status])  
                if status is COM_STATUS.LOGIN_READY:
                    pkg = self.__RcvMessageQueue.get(timeout=5)
                    # print(datetime.datetime.now(), "DT=[{dt}] {Log}".format(dt=pkg.DT,Log=pkg.ToLog())) 
                    P1503 = P001503(pkg)
                    if (P1503.Code != 0):
                        continue
                    for _p1503 in P1503.p001503_2:
                        self.accounts.append(_p1503)
                else:
                    print('TradeCom登入失敗' , status, self.__Com_StatusDict.get(status))
        except Empty:
            print('TradeCom登入失敗 Timeout' ,self.__com_status, self.__Com_StatusDict.get(self.__com_status))
        
        # import time
        # for acc in self.accounts:
        #     topic = f'{acc.BrokeId}-{acc.Account}'
        #     while topic not in self.RecoverTopic:
        #         time.sleep(0.01)
        if len(self.accounts) > 0 : print("TradeCom登入成功")
        return self.accounts

    def Login_RcvMessage(self, dt:DT, pkg):
        if dt == DT.LOGIN:
            self.__RcvMessageQueue.put(pkg)
            return True
        else:
            return False
        
    def Login_GetStatus(self, status:COM_STATUS, msg):
        if status in [COM_STATUS.LOGIN_READY, COM_STATUS.CONNECT_FAIL, COM_STATUS.NOVALIDCERT, COM_STATUS.LOGIN_FAIL, COM_STATUS.DISCONNECTED, COM_STATUS.NOCGCSPIAPI]:
            self.__com_status = status
            self.__Com_StatusDict[self.__com_status] = { 'Time' : datetime.datetime.now() , 'Msg' : msg }
            self.__GetStatusQueue.put(status)


from setuptools import setup, find_packages

with open('README.md', 'r', encoding='UTF-8') as f:
    long_description = f.read()

setup(
    name='kgisuperpy', #pip download project name
    version='1.0.2',
    description='KGI project',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='KGI Securities',
    author_email='ec.service@kgi.com',
    packages=find_packages(),
    package_data={
        'superpy': [    #dir name and import namespace
            './.libs/Package.dll',
            './.libs/PushClient.dll',
            './.libs/TradeCom.dll',
            './.libs/QuoteCom.dll',
            './.libs/ICSharpCode.SharpZipLib.dll',
            './.libs/Interop.KGICGCAPIATLLib.dll'
        ]
    },
    install_requires=[
        'pythonnet'
    ]
)


import sys, os
import clr
from enum import Enum

assembly_path = os.path.dirname(__file__)
sys.path.append(assembly_path)
clr.AddReference("Package")     #必要引用dll

from Intelligence import COM_STATUS
from Intelligence import RECOVER_STATUS

from Intelligence import Security_OrdType
from Intelligence import Security_Lot
from Intelligence import Security_Class
from Intelligence import Security_PriceFlag
from Intelligence import SIDE_FLAG
from Intelligence import TIME_IN_FORCE
from Intelligence import ORDER_TYPE
from Intelligence import MARKET_FLAG
from Intelligence import PRICE_FLAG
from Intelligence import POSITION_EFFECT
from Intelligence import OFFICE_FLAG

from DT import *

class COM_STATUS(Enum):    
    CONNECT_READY = COM_STATUS.CONNECT_READY
    CONNECT_FAIL = COM_STATUS.CONNECT_FAIL
    DISCONNECTED = COM_STATUS.DISCONNECTED
    LOGIN_READY = COM_STATUS.LOGIN_READY
    LOGIN_FAIL = COM_STATUS.LOGIN_FAIL
    LOGIN_UNKNOW = COM_STATUS.LOGIN_UNKNOW
    SUBSCRIBE = COM_STATUS.SUBSCRIBE
    UNSUBSCRIBE = COM_STATUS.UNSUBSCRIBE
    HEART_BEAT = COM_STATUS.HEART_BEAT
    ACK_REQUESTID = COM_STATUS.ACK_REQUESTID
    RECOVER_DATA = COM_STATUS.RECOVER_DATA
    LOGIN_FROM_CLOSE = COM_STATUS.LOGIN_FROM_CLOSE
    AS400_CONNECTED = COM_STATUS.AS400_CONNECTED
    AS400_CONNECTFAIL = COM_STATUS.AS400_CONNECTFAIL
    AS400_DISCONNECTED = COM_STATUS.AS400_DISCONNECTED
    QUEUE_WARNING = COM_STATUS.QUEUE_WARNING
    ACCTCNT_NOMATCH = COM_STATUS.ACCTCNT_NOMATCH
    NOVALIDCERT = COM_STATUS.NOVALIDCERT
    NOCGCSPIAPI = COM_STATUS.NOCGCSPIAPI

class RECOVER_STATUS(Enum):
    RS_BEGIN = RECOVER_STATUS.RS_BEGIN
    RS_DONE = RECOVER_STATUS.RS_DONE
    RS_NOAUTHRITY = RECOVER_STATUS.RS_NOAUTHRITY

class Security_OrdType:
    OT_NEW = Security_OrdType.OT_NEW
    OT_CANCEL = Security_OrdType.OT_CANCEL
    OT_MODIFY_QTY = Security_OrdType.OT_MODIFY_QTY
    OT_MODIFY_PRICE = Security_OrdType.OT_MODIFY_PRICE

class Security_Lot:
    Even_Lot = Security_Lot.Even_Lot
    Odd_Lot = Security_Lot.Odd_Lot
    Fixed_Price = Security_Lot.Fixed_Price
    Block_Trade = Security_Lot.Block_Trade
    Odd_InTraday = Security_Lot.Odd_InTraday

class Security_Class:
    SC_Ordinary = Security_Class.SC_Ordinary
    SC_SelfMargin = Security_Class.SC_SelfMargin
    SC_SelfShort = Security_Class.SC_SelfShort
    SC_ShortLimit = Security_Class.SC_ShortLimit
    SC_ShortUnLimit = Security_Class.SC_ShortUnLimit
    SC_DayMargin = Security_Class.SC_DayMargin
    SC_DayShort = Security_Class.SC_DayShort
    SC_DayTrade = Security_Class.SC_DayTrade
    SC_EDN = Security_Class.SC_EDN

class Security_PriceFlag:
    SP_FixedPrice = Security_PriceFlag.SP_FixedPrice
    SP_FallStopPrice = Security_PriceFlag.SP_FallStopPrice
    SP_UnchangePrice = Security_PriceFlag.SP_UnchangePrice
    SP_RiseStopPrice = Security_PriceFlag.SP_RiseStopPrice
    SP_MarketPrice = Security_PriceFlag.SP_MarketPrice

class SIDE_FLAG:
    SF_BUY = SIDE_FLAG.SF_BUY
    SF_SELL = SIDE_FLAG.SF_SELL
    
class TIME_IN_FORCE:
    TIF_ROD = TIME_IN_FORCE.TIF_ROD
    TIF_IOC = TIME_IN_FORCE.TIF_IOC
    TIF_FOK = TIME_IN_FORCE.TIF_FOK

class ORDER_TYPE:
    OT_NEW = ORDER_TYPE.OT_NEW
    OT_CANCEL = ORDER_TYPE.OT_CANCEL
    OT_MODIFY_QTY = ORDER_TYPE.OT_MODIFY_QTY
    OT_MODIFY_PRICE = ORDER_TYPE.OT_MODIFY_PRICE
    OT_MODIFY = ORDER_TYPE.OT_MODIFY

class MARKET_FLAG:
    MF_FUT = MARKET_FLAG.MF_FUT
    MF_OPT = MARKET_FLAG.MF_OPT
    MF_STK = MARKET_FLAG.MF_STK
    MF_FUT_SPREAD = MARKET_FLAG.MF_FUT_SPREAD
    MF_OPT_SPREAD = MARKET_FLAG.MF_OPT_SPREAD

class PRICE_FLAG:
    PF_SPECIFIED = PRICE_FLAG.PF_SPECIFIED
    PF_MARKET = PRICE_FLAG.PF_MARKET
    PF_STOP_MARKET = PRICE_FLAG.PF_STOP_MARKET
    PF_STOP_SPECIFIED = PRICE_FLAG.PF_STOP_SPECIFIED
    PF_MARKET_RANGE = PRICE_FLAG.PF_MARKET_RANGE

class POSITION_EFFECT:
    PE_OPEN = POSITION_EFFECT.PE_OPEN
    PE_CLOSE = POSITION_EFFECT.PE_CLOSE
    PE_DAY_TRADE = POSITION_EFFECT.PE_DAY_TRADE
    PE_AUTO = POSITION_EFFECT.PE_AUTO

class OFFICE_FLAG:
    OF_SPEEDY = OFFICE_FLAG.OF_SPEEDY
    OF_AS400 = OFFICE_FLAG.OF_AS400


from queue import Queue, Empty
from System import Decimal as CSDecimal
from datetime import datetime

from IntelligencePy import ORDER_TYPE
from IntelligencePy import MARKET_FLAG
from IntelligencePy import SIDE_FLAG
from IntelligencePy import PRICE_FLAG
from IntelligencePy import TIME_IN_FORCE
from IntelligencePy import POSITION_EFFECT
from IntelligencePy import OFFICE_FLAG

from PackagePy import PT02010, PT02002

class Order:
    def __init__(self, RequestIds = dict(), callback = lambda dt, dic: print(dt, dic)):
        self.tradecom = None
        self.__RequestIdDict = RequestIds
        self.callback = callback

    def FutOrder(self, type: ORDER_TYPE, market: MARKET_FLAG, 
                 brokerId, account, subAccount, symbolId, 
                 BS: SIDE_FLAG, pricefl: PRICE_FLAG, price, TIF: TIME_IN_FORCE
              , qty, pf: POSITION_EFFECT, off: OFFICE_FLAG, webid = '', cnt= '', orderno= '') -> PT02010:
        

        QTY = int(qty)
        PRICE = CSDecimal(float(price))
        
        RequestId = self.tradecom.GetRequestId()
        print(f"送單 RequestId=[{RequestId}]")
        if type == ORDER_TYPE.OT_NEW:
            res = self.tradecom.Order(ORDER_TYPE.OT_NEW, market, RequestId, brokerId, account, subAccount, symbolId, BS, pricefl, PRICE, TIF, QTY, pf, OFFICE_FLAG.OF_AS400)
        else:
            res = self.tradecom.Order(type, market, RequestId, brokerId, account, subAccount, symbolId, BS, pricefl, PRICE, TIF, QTY, pf, OFFICE_FLAG.OF_AS400, webid, cnt, orderno)
        if res != 0:
            print("委託失敗: ", self.tradecom.GetOrderErrMsg(res))
        else:
            print("委託成功: ")

        self.__RequestIdDict[RequestId] = Queue()
        FUT_ORDER_RPT = self.getPkg(self.__RequestIdDict[RequestId])
        if isinstance(FUT_ORDER_RPT, PT02010):
            return PT02010(FUT_ORDER_RPT)
        elif isinstance(FUT_ORDER_RPT, PT02002):
            FUT_ORDER_RPT.OrderNo = ""
            FUT_ORDER_RPT.Code = FUT_ORDER_RPT.ErrorCode
            FUT_ORDER_RPT.TradeDate = datetime.now().strftime("%Y%m%d")
            FUT_ORDER_RPT.ReportTime = datetime.now().strftime("%H%M%S%f")
            FUT_ORDER_RPT.AfterQty = 0
            FUT_ORDER_RPT.WebID = FUT_ORDER_RPT.WEBID
            return FUT_ORDER_RPT
        else:
            return None